# # -*- coding: utf-8 -*-
# import os
#
# from frm.enums.term_structures import FXSmileInterpolationMethod
#
# if __name__ == "__main__":
#     os.chdir(os.environ.get('PROJECT_DIR_FRM'))
#
# from frm.term_structures.fx_volatility_surface_helpers import interp_fx_forward_curve_df
# from frm.term_structures.zero_curve import ZeroCurve
# from frm.pricing_engine.monte_carlo_generic import generate_rand_nbs
# from frm.pricing_engine.geometric_brownian_motion import simulate_gbm_path
# from frm.pricing_engine.heston import simulate_heston
#
# from frm.utils.daycount import year_fraction
#
# from frm.enums.utils import DayCountBasis
#
# import numpy as np
# import pandas as pd
# from dataclasses import dataclass, field
# from typing import Optional
# import matplotlib.pyplot as plt
#
#
# # Models TBC: sabr, svi_sabr, vanna_volga, sv_local_vol, jump_diffusion
#
#
#
#
#
# @dataclass
# class FXVolatilitySurface:
#     # Mandatory initialisation attributes
#     curve_date: pd.Timestamp
#     spot_date: pd.Timestamp
#     curve_ccy: str
#     fx_forward_curve: pd.DataFrame
#     domestic_zero_curve: ZeroCurve
#     foreign_zero_curve: ZeroCurve
#
#
#     # Optional initialisation attributes
#     spot_date: Optional[pd.Timestamp] = None
#     day_count_basis: DayCountBasis = DayCountBasis.ACT_ACT
#     σ_pillar: pd.DataFrame=None
#     smile_interpolation_method: FXSmileInterpolationMethod = FXSmileInterpolationMethod.HESTON_COSINE
#
#     # Non initialisation attributes set in __post_init__
#     fx_spot_rate: float = field(init=False)
#     domestic_ccy: str = field(init=False)
#     foreign_ccy: str = field(init=False)
#     daily_volatility_smile: pd.DataFrame=None
#     σ_pillar: dict=None
#     daily_volatility_smile_func: dict=None
#
#     # volatility definitions
#
#
#
#
#
#
#     # def __post_init__(self):
#     #
#     #     self.daily_volatility_smile_func = dict()
#     #
#     #     busdaycal = get_busdaycal(keys=[self.curve_ccy[:3],self.curve_ccy[3:]])
#     #
#     #     if self.spot_date is None:
#     #         result = get_tenor_settlement_date(self.curve_date, 'sp', self.curve_ccy, busdaycal=busdaycal, spot_offset=True)
#     #         holiday_rolled_offset_date, cleaned_tenor, spot_date = result
#     #         assert holiday_rolled_offset_date == spot_date
#     #         self.spot_date = spot_date
#     #
#     #     # curve_ccy validation
#     #     self.curve_ccy = self.curve_ccy.lower().strip()
#     #     assert len(self.curve_ccy) == 6, self.curve_ccy
#     #     self.foreign_ccy = self.curve_ccy[:3]
#     #     self.domestic_ccy = self.curve_ccy[3:]
#     #     if self.foreign_ccy == 'usd' and self.domestic_ccy in {'aud','eur','gbp','nzd'}:
#     #         warnings.warn("non conventional fx market, market convention is 'usd' as the domestic currency for 'aud/usd', 'eur/usd', 'gbp/usd' and 'nzd/usd' pairs")
#     #     elif self.domestic_ccy == 'usd' and self.foreign_ccy not in {'aud','eur','gbp','nzd'}:
#     #         warnings.warn("non conventional fx market, market convention is for 'usd' to be the foreign currency (except for 'aud/usd', 'eur/usd', 'gbp/usd' and 'nzd/usd' pairs)")
#     #
#     #     # fx_forward_curve validation
#     #     assert 'fx_forward_rate' in self.fx_forward_curve.columns
#     #     self.fx_forward_curve = convert_column_to_consistent_data_type(self.fx_forward_curve)
#     #     assert ('tenor_name' in self.fx_forward_curve.columns) or ('expiry_date' in self.fx_forward_curve.columns and 'delivery_date' in self.fx_forward_curve.columns)
#     #     if 'expiry_date' not in self.fx_forward_curve.columns and 'delivery_date' not in self.fx_forward_curve.columns:
#     #
#     #         # Calculate and set the delivery date
#     #         settlement_date, cleaned_tenor, spot_date = get_tenor_settlement_date(self.curve_date, self.fx_forward_curve['tenor_name'], self.curve_ccy, busdaycal=busdaycal, spot_offset=True)
#     #         self.fx_forward_curve['tenor'] = cleaned_tenor
#     #         self.fx_forward_curve['delivery_date'] = settlement_date
#     #         self.fx_forward_curve['years_to_delivery'] = year_fraction(self.curve_date, self.fx_forward_curve['delivery_date'], self.day_count_basis)
#     #
#     #         # Calculate and set the expiry date
#     #         spot_offset = -1 * get_spot_offset(curve_ccy=self.curve_ccy)
#     #         expiry_date = np.busday_offset(settlement_date.values.astype('datetime64[D]'), offsets=spot_offset, roll='preceding', busdaycal=busdaycal)
#     #         self.fx_forward_curve['expiry_date'] = expiry_date
#     #         self.fx_forward_curve['years_to_expiry'] = year_fraction(self.curve_date, self.fx_forward_curve['expiry_date'], self.day_count_basis)
#     #
#     #     if self.curve_date not in self.fx_forward_curve['expiry_date'].values:
#     #         raise ValueError("The curve date (" + self.curve_date.strftime('%Y-%m-%d') + ') fx rate (i.e the fx spot rate) is in missing in fx_forward_curve')
#     #     else:
#     #         mask = self.fx_forward_curve['expiry_date'] == self.curve_date
#     #         self.fx_spot_rate = self.fx_forward_curve.loc[mask,'fx_forward_rate'].iloc[0]
#     #
#     #     # σ_pillar validation
#     #     if self.σ_pillar is not None:
#     #         σ_pillar = self.σ_pillar
#     #         σ_pillar['curve_date'] = self.curve_date
#     #         σ_pillar['curve_ccy'] = self.curve_ccy
#     #         σ_pillar['day_count_basis'] = self.day_count_basis.value
#     #
#     #         σ_pillar = generic_market_data_input_cleanup_and_validation(σ_pillar, spot_offset=False)
#     #         σ_pillar = fx_σ_input_helper(σ_pillar)
#     #         σ_pillar.set_index('tenor_date', drop=True, inplace=True)
#     #
#     #         σ_pillar = self.__interp_daily_FXF_and_IR_rates(σ_pillar)
#     #
#     #         self.σ_pillar = σ_pillar
#     #
#     #         if 'σ_atmf' in self.σ_pillar.keys():
#     #             self.K_pillar = σ_pillar['fx_forward_rate']
#     #         else:
#     #             ###################################################################
#     #             # Calculate the delta-strike surface from the delta-volatility surface
#     #             K_pillar = self.σ_pillar.copy()
#     #             K_pillar.loc[:, K_pillar.columns.str.contains('σ')] = np.nan
#     #             K_pillar.rename(columns=lambda x: x.replace('σ', 'k'), inplace=True)
#     #
#     #             dict_K_σ = {v.replace('σ', 'k'): v for v in self.σ_pillar.columns if 'σ' in v}
#     #             for date, row in K_pillar.iterrows():
#     #                 # Scalars
#     #                 S0 = self.fx_spot_rate
#     #                 r_f=row['foreign_ccy_continuously_compounded_zero_rate']
#     #                 r_d=row['domestic_ccy_continuously_compounded_zero_rate']
#     #                 tau=row['tenor_years']
#     #                 delta_convention=row['Δ_convention']
#     #                 F=row['fx_forward_rate']
#     #                 # Arrays
#     #                 cp = [1 if v[-4:] == 'call' else -1 if v[-3:] == 'put' else None for v in dict_K_σ.keys()]
#     #                 delta = [0.5 if v == 'k_atmΔneutral' else cp[i] * float(v.split('_')[1].split('Δ')[0]) / 100 for i,v in enumerate(dict_K_σ.keys())]
#     #                 vol = self.σ_pillar.loc[date, dict_K_σ.values()].values
#     #
#     #                 K = gk_solve_strike(S0=S0,tau=tau,r_f=r_f,r_d=r_d,σ=σ,delta=delta,delta_convention=delta_convention,F=F)
#     #                 K_pillar.loc[date,dict_K_σ.keys()] = K
#     #
#     #             self.K_pillar = K_pillar
#     #
#     #         ###################################################################
#     #         # Interpolate the term structure of volatility to get a daily volatility smile
#     #         dates = pd.date_range(min(self.σ_pillar.index),max(self.σ_pillar.index),freq='d')
#     #         tenor_years = year_fraction(self.curve_date,dates, self.day_count_basis)
#     #         df_interp = pd.DataFrame({'tenor_date':dates, 'tenor_years': tenor_years})
#     #         df_pillar = self.σ_pillar.copy()
#     #
#     #         # Merge to find closest smaller and larger tenors for each target tenor
#     #         df_lower_pillar = pd.merge_asof(df_interp.sort_values('tenor_years'), df_pillar, on='date', direction='backward', suffixes=('_interp', '_pillar'))
#     #         df_upper_pillar = pd.merge_asof(df_interp.sort_values('tenor_years'), df_pillar, on='date', direction='forward', suffixes=('_interp', '_pillar'))
#     #
#     #         # cols equals the volatility smile to be interpolated
#     #         cols = [v for v in df_pillar.columns.to_list() if 'σ' in v]
#     #
#     #         # Convert to numpy for efficient calculations
#     #         t1 = df_lower_pillar['tenor_years_pillar'].to_numpy()
#     #         t2 = df_upper_pillar['tenor_years_pillar'].to_numpy()
#     #         t = df_interp['tenor_years'].to_numpy()
#     #         t1 = t1[:, np.newaxis]
#     #         t2 = t2[:, np.newaxis]
#     #         t = t[:, np.newaxis]
#     #         σ_t1 = df_lower_pillar[cols].to_numpy()
#     #         σ_t2 = df_upper_pillar[cols].to_numpy()
#     #
#     #         # Interpolation logic
#     #         mask = t1 != t2
#     #         σ_t = np.where(mask, self.flat_forward_interp(t1, σ_t1, t2, σ_t2, t), σ_t1)
#     #         df_interp[cols] = σ_t
#     #         df_interp['Δ_convention'] = 'regular_forward_Δ' # need to add a section to calculate forward delta from spot delta
#     #         df_interp.set_index('tenor_date', inplace=True,drop=True)
#     #
#     #         self.daily_volatility_smile = self.__interp_daily_FXF_and_IR_rates(df_interp)
#
#     #
#     # def __interp_daily_FXF_and_IR_rates(self, df):
#     #     # to do - need to rework so index is 0,1,2,3... and
#     #     # 'expiry_date' and 'delivery_date' are two new columns.
#     #
#     #     df['foreign_ccy_continuously_compounded_zero_rate'] = np.nan
#     #     df['domestic_ccy_continuously_compounded_zero_rate'] = np.nan
#     #     df['fx_forward_rate'] = np.nan
#     #
#     #     df['foreign_ccy_continuously_compounded_zero_rate'] = self.foreign_zero_curve.zero_rate(dates=df.index,compounding_frequency=CompoundingFrequency.CONTINUOUS).values
#     #     df['domestic_ccy_continuously_compounded_zero_rate'] = self.domestic_zero_curve.zero_rate(dates=df.index,compounding_frequency=CompoundingFrequency.CONTINUOUS).values
#     #     df['fx_forward_rate'] = interp_fx_forward_curve(self.fx_forward_curve, dates=df.index, date_type='expiry_date', flat_extrapolation=True).values
#     #     return df
#     #
#     #
#     # def forward_volatility(self,
#     #                        t1: Union[float, np.array],
#     #                        σ_t1: Union[float, np.array],
#     #                        t2: Union[float, np.array],
#     #                        σ_t2: Union[float, np.array]) -> Union[float, np.array]:
#     #     """
#     #     Calculate the at-the-money forward volatility from time t1 to t2.
#     #     The forward volatility is based on the consistency condition:
#     #     σ_t1**2 * t1 + σ_t1_t2**2 * (t2- t1) = σ_t1**2 * t2
#     #
#     #     Parameters:
#     #     - t1 (float): Time to first maturity
#     #     - σ_t1 (float): At-the-money volatility at time (in years) to expiry date 1
#     #     - t2 (float): Time to second maturity
#     #     - σ_t2 (float): At-the-money volatility at time (in years) to expiry date 2
#     #
#     #     Returns:
#     #     - np.array: Forward volatility from time t1 to t2
#     #     """
#     #     tau = t2 - t1
#     #     if np.any(tau == 0):
#     #         warnings.warn("t2 and t1 are equal. NaN will be returned.")
#     #     elif np.any(tau < 0):
#     #         raise ValueError("t2 is less than t1. Please swap ")
#     #
#     #     result = (σ_t2**2 * t2 - σ_t1**2 * t1) / tau
#     #     if np.any(result < 0):
#     #         raise ValueError("Negative value encountered under square root.")
#     #
#     #     return np.sqrt(result)
#     #
#     # def flat_forward_interp(self,
#     #                         t1: Union[float, np.array],
#     #                         σ_t1: Union[float, np.array],
#     #                         t2: Union[float, np.array],
#     #                         σ_t2: Union[float, np.array],
#     #                         t: Union[float, np.array]) -> Union[float, np.array]:
#     #     """
#     #     Interpolate the at-the-money volatility at a given time 't' using flat forward interpolation.
#     #
#     #     Parameters:
#     #     - t1 (Union[float, array]): Time to first maturity
#     #     - σ_t1 (Union[float, array]): At-the-money volatility at time t1
#     #     - t2 (Union[float, array]): Time to second maturity
#     #     - σ_t2 (Union[float, array]): At-the-money volatility at time t2
#     #     - t (Union[float, array]): Time at which to interpolate the volatility
#     #
#     #     Returns:
#     #     - array: Interpolated volatility at time 't'
#     #     """
#     #     σ_t12 = self.forward_volatility(t1, σ_t1, t2, σ_t2)
#     #     return np.sqrt((σ_t1**2 * t1 + σ_t12**2 * (t - t1)) / t)
#
#
#
#     def fit_and_plot_smile(self):
#
#         def convert_column_name_to_delta_for_plt(input_str):
#             if 'put' in input_str:
#                 return int(input_str.split('Δ')[0][2:])
#             elif 'call' in input_str:
#                 return 100 - int(input_str.split('Δ')[0][2:])
#             elif 'neutral' in input_str:
#                 return 50
#
#         def convert_column_name_to_delta(input_str):
#             if 'put' in input_str:
#                 return -1 * int(input_str.split('Δ')[0][2:]) / 100
#             elif 'call' in input_str:
#                 return int(input_str.split('Δ')[0][2:]) / 100
#             elif 'neutral' in input_str:
#                 return 0.5
#
#         i = 0
#         for date, row in self.σ_pillar.iterrows():
#
#             # Scalars
#             tenor_name = row['tenor_name']
#             Δ_convention=row['Δ_convention']
#             S = self.fx_spot_rate
#             r_f=row['foreign_ccy_continuously_compounded_zero_rate']
#             r_d=row['domestic_ccy_continuously_compounded_zero_rate']
#             tau=row['tenor_years']
#             F=row['fx_forward_rate']
#
#             if F is not None:
#                 # Use market forward rate and imply the curry basis-adjusted domestic interest rate
#                 F = np.atleast_1d(F).astype(float)
#                 r_d_basis_adj = np.log(F / S) / tau + r_f # from F = S0 * exp((r_d - r_f) * tau)
#                 r = r_d_basis_adj
#                 q = r_f
#             else:
#                 # By interest rate parity
#                 F = S * np.exp((r_d - r_f) * tau)
#                 r = r_d  # noqa: F841
#                 q = r_f  # noqa: F841
#
#             # Arrays
#             Δ = [convert_column_name_to_delta(c) for c in self.σ_pillar.columns if c[0] == 'σ']
#             Δ_for_plt = [convert_column_name_to_delta_for_plt(c) for c in self.σ_pillar.columns if c[0] == 'σ']
#             cp = np.sign(Δ)
#             σ = row[[c for c in self.σ_pillar.columns if c[0] == 'σ']].values
#
#             if self.smile_interpolation_method[:6] == 'heston':
#                 var0, vv, kappa, theta, rho, lambda_, IV, SSE = \
#                     heston_calibrate_vanilla_smile(Δ, Δ_convention, σ, S, r_f, r_d, tau, cp, pricing_method=self.smile_interpolation_method)
#
#                 # Displaying output
#                 print(f'=== {tenor_name} calibration results ===')
#                 print(f'var0, vv, kappa, theta, rho: {var0, vv, kappa, theta, rho}')
#                 print(f'SSE {SSE*100}')
#
#                 # Plotting
#                 i+=1
#                 plt.figure(i+1)
#                 plt.plot(Δ_for_plt, σ * 100, 'ko-', linewidth=1)
#                 plt.plot(Δ_for_plt, IV * 100, 'rs--', linewidth=1)
#                 plt.legend([f'{tenor_name} smile', 'Heston fit'], loc='upper right')
#                 plt.xlabel('Delta [%]')
#                 plt.ylabel('Implied volatility [%]')
#                 plt.xticks(Δ_for_plt)
#                 plt.title(self.smile_interpolation_method)
#                 plt.show()
#
#             else:
#                 raise ValueError
#
#
#
#
#
#     # def interp_σ_smile_from_pillar_points(self,
#     #                      expiry_dates: pd.DatetimeIndex):
#     #     """
#     #     Private method to interpolate the σ surface for given dates, updating the
#     #     `daily_volatility_smile_func` attribute with appropriate interpolation functions or Heston parameters.
#     #
#     #     Parameters:
#     #     dates (pd.DatetimeIndex): Array of dates to perform the σ surface interpolation for.
#     #
#     #     Raises:
#     #     ValueError: If the sum of squared errors (SSE) from the Heston fit exceeds a threshold, indicating a poor fit.
#     #
#     #     Note:
#     #     Updates the `daily_volatility_smile_func` attribute with either a spline interpolation function or Heston parameters for each date in `dates`.
#     #     """
#     #
#     #     df_unique_dates = self.daily_volatility_smile.loc[pd.DatetimeIndex(set(expiry_dates)),:].copy()
#     #     df_unique_dates.loc[:, df_unique_dates.columns.str.contains('σ')] = np.nan
#     #     df_unique_dates.rename(columns=lambda x: x.replace('σ', 'k'), inplace=True)
#     #
#     #     helper_cols = ['tenor_date',
#     #                    'tenor_name',
#     #                    'tenor_years',
#     #                    'fx_forward_rate',
#     #                    'foreign_ccy_continuously_compounded_zero_rate',
#     #                    'domestic_ccy_continuously_compounded_zero_rate',
#     #                    'Δ_convention']
#     #
#     #     dict_K_σ = {v.replace('σ', 'k'): v for v in self.daily_volatility_smile.columns if v not in helper_cols}
#     #
#     #     # Interpolate volatility surface for each date in df
#     #     for expiry_date, row in df_unique_dates.iterrows():
#     #
#     #         if expiry_date not in self.daily_volatility_smile_func.keys():
#     #
#     #             # Scalars
#     #             S0 = self.fx_spot_rate
#     #             r_f=row['foreign_ccy_continuously_compounded_zero_rate']
#     #             r_d=row['domestic_ccy_continuously_compounded_zero_rate']
#     #             tau=row['tenor_years']
#     #             delta_convention=row['Δ_convention']
#     #             F=row['fx_forward_rate']
#     #
#     #             # Arrays
#     #             cp = [1 if v[-4:] == 'call' else -1 if v[-3:] == 'put' else 1 for v in dict_K_σ.keys()]
#     #             delta = [0.5 if v == 'k_atmΔneutral' else cp[i] * float(v.split('_')[1].split('Δ')[0]) / 100 for i,v in enumerate(dict_K_σ.keys())]
#     #             vol = self.daily_volatility_smile.loc[expiry_date, dict_K_σ.values()].values
#     #
#     #             if self.smile_interpolation_method in [FXSmileInterpolationMethod.UNIVARIATE_SPLINE, FXSmileInterpolationMethod.CUBIC_SPLINE]:
#     #                 K = gk_solve_strike(S0=S0,tau=tau,r_d=r_d,r_f=r_f,vol=vol,delta=delta,delta_convention=delta_convention,F=F)
#     #                 df_unique_dates.loc[expiry_date,dict_K_σ.keys()] = K
#     #                 if self.smile_interpolation_method == FXSmileInterpolationMethod.UNIVARIATE_SPLINE
#     #                     self.daily_volatility_smile_func[expiry_date] = InterpolatedUnivariateSpline(x=K, y=vol)
#     #                 elif self.smile_interpolation_method == FXSmileInterpolationMethod.CUBIC_SPLINE:
#     #                     self.daily_volatility_smile_func[expiry_date] = CubicSpline(x=K, y=vol)
#     #             elif self.smile_interpolation_method in [FXSmileInterpolationMethod.HESTON_ANALYTICAL_1993,
#     #                                                      FXSmileInterpolationMethod.HESTON_CARR_MADAN_GAUSS_KRONROD_QUADRATURE,
#     #                                                      FXSmileInterpolationMethod.HESTON_CARR_MADAN_FFT_W_SIMPSONS,
#     #                                                      FXSmileInterpolationMethod.HESTON_COSINE,
#     #                                                      FXSmileInterpolationMethod.HESTON_LIPTON]:
#     #                 var0, vv, kappa, theta, rho, lambda_, IV, SSE = \
#     #                     heston_calibrate_vanilla_smile(delta=delta, delta_convention, σ, S0, r_d, r_f, tau, cp, pricing_method=self.smile_interpolation_method)
#     #                 if SSE < 0.001:
#     #                     result = {
#     #                         'var0': var0,
#     #                         'vv': vv,
#     #                         'kappa': kappa,
#     #                         'theta': theta,
#     #                         'rho': rho,
#     #                         'lambda_': lambda_,
#     #                         'IV': IV,
#     #                         'SSE': SSE
#     #                         }
#     #
#     #                     self.daily_volatility_smile_func[expiry_date] = result
#     #                 else:
#     #                     raise ValueError('SSE is a large value, ', round(SSE,4), ' heston fit at ', expiry_date,' is likely poor')
#
#
#     # def interp_σ_surface(self,
#     #                       expiry_dates: pd.DatetimeIndex,
#     #                       K: np.array,
#     #                       cp: np.array):
#     #
#     #     K = np.atleast_1d(K).astype(float)
#     #     cp = np.atleast_1d(cp).astype(float)
#     #
#     #     assert expiry_dates.shape == K.shape
#     #     assert expiry_dates.shape == cp.shape
#     #
#     #     # Interpolate the volatility smile for the given expiries
#     #     self.interp_σ_smile_from_pillar_points(expiry_dates)
#     #
#     #
#     #     df = self.daily_volatility_smile.loc[expiry_dates,:].copy()
#     #
#     #     cols_to_drop = [col for col in df.columns if 'σ' in col]
#     #     df.drop(columns=cols_to_drop, inplace=True)
#     #     df['K'] = K
#     #     df['call_put'] = cp
#     #
#     #     #helper_cols = ['tenor_date','tenor_name','tenor_years','fx_forward_rate','foreign_ccy_continuously_compounded_zero_rate','domestic_ccy_continuously_compounded_zero_rate','Δ_convention']
#     #     # dict_K_σ = {v.replace('σ', 'k'): v for v in self.daily_volatility_smile.columns if v not in helper_cols}
#     #
#     #     result = []
#     #
#     #     # Get the implied vol for each strike and date combination
#     #     # At some point need to make this over unique date, K
#     #     for expiry_date,row in df.iterrows():
#     #
#     #         K_target = row['K']
#     #
#     #         if self.smile_interpolation_method in {'univariate_spline', 'cubic_spline'}:
#     #
#     #             σ = self.daily_volatility_smile_func[expiry_date](K_target)
#     #             result.append(σ)
#     #
#     #         elif self.smile_interpolation_method[:6] == 'heston':
#     #
#     #             S0 = self.fx_spot_rate
#     #             r_f = row['foreign_ccy_continuously_compounded_zero_rate']
#     #             r_d = row['domestic_ccy_continuously_compounded_zero_rate']
#     #             tau = row['tenor_years']
#     #             F = row['fx_forward_rate']
#     #             cp = row['call_put']
#     #
#     #             if F is not None:
#     #                 # Use market forward rate and imply the curry basis-adjusted domestic interest rate
#     #                 F = np.atleast_1d(F).astype(float)
#     #                 r_d_basis_adj = np.log(F / S0) / tau + r_f # from F = S0 * exp((r_d - r_f) * tau)
#     #                 r = r_d_basis_adj
#     #                 q = r_f
#     #             else:
#     #                 # By interest rate parity
#     #                 F = S0 * np.exp((r_d - r_f) * tau)
#     #                 r = r_d
#     #                 q = r_f
#     #
#     #             var0 = self.daily_volatility_smile_func[expiry_date]['var0']
#     #             vv = self.daily_volatility_smile_func[expiry_date]['vv']
#     #             kappa = self.daily_volatility_smile_func[expiry_date]['kappa']
#     #             theta = self.daily_volatility_smile_func[expiry_date]['theta']
#     #             rho = self.daily_volatility_smile_func[expiry_date]['rho']
#     #             lambda_ = self.daily_volatility_smile_func[expiry_date]['lambda_']
#     #
#     #             if self.smile_interpolation_method == 'heston_analytical_1993':
#     #                 X = heston1993_price_vanilla_european(S0=S0, tau=tau, r=r, q=q, cp=cp, K=K_target, var0=var0, vv=vv, kappa=kappa, theta=theta, rho=rho, lambda_=lambda_)
#     #             elif self.smile_interpolation_method == 'heston_carr_madan_gauss_kronrod_quadrature':
#     #                 X = heston_carr_madan_price_vanilla_european(S0=S0, tau=tau, r=r, q=q, cp=cp, K=K_target, var0=var0, vv=vv, kappa=kappa, theta=theta, rho=rho, integration_method=0)
#     #             elif self.smile_interpolation_method == 'heston_carr_madan_fft_w_simpsons':
#     #                 X = heston_carr_madan_price_vanilla_european(S0=S0, tau=tau, r=r, q=q, cp=cp, K=K_target, var0=var0, vv=vv, kappa=kappa, theta=theta, rho=rho, integration_method=1)
#     #             elif self.smile_interpolation_method == 'heston_cosine':
#     #                 X = heston_cosine_price_vanilla_european(S0=S0, tau=tau, r=r, q=q, cp=cp, K=K_target, var0=var0, vv=vv, kappa=kappa, theta=theta, rho=rho)
#     #             elif self.smile_interpolation_method == 'heston_lipton':
#     #                 X = heston_lipton_price_vanilla_european(S0=S0, tau=tau, r=r, q=q, cp=cp, K=K_target, var0=var0, vv=vv, kappa=kappa, theta=theta, rho=rho)
#     #
#     #             result.append(gk_solve_implied_volatility(S0=S0, tau=tau, r_d=r, r_f=q, cp=cp, K=K_target, X=X, vol_guess=var0**2))
#     #     return np.array(result)
#
#
#     # def price_vanilla_european(self,
#     #                            expiry_datetimeindex: pd.DatetimeIndex,
#     #                            K,
#     #                            cp,
#     #                            analytical_greeks_flag,
#     #                            intrinsic_time_split_flag):
#     #
#     #     K = np.atleast_1d(K)
#     #     cp = np.atleast_1d(cp)
#     #
#     #     assert expiry_datetimeindex.shape == K.shape
#     #     assert expiry_datetimeindex.shape == cp.shape
#     #
#     #     mask = np.logical_and.reduce([expiry_datetimeindex >= self.σ_pillar.index.min(), expiry_datetimeindex <= self.σ_pillar.index.max()])
#     #     vol = np.full(expiry_datetimeindex.shape, np.nan)
#     #
#     #     if mask.any():
#     #         self.interp_σ_smile_from_pillar_points(expiry_datetimeindex[mask])
#     #         σ[mask] = self.interp_σ_surface(expiry_datetimeindex[mask], K, cp)
#     #
#     #     r_d = self.domestic_zero_curve.zero_rate(dates=expiry_datetimeindex,compounding_frequency=CompoundingFrequency.CONTINUOUS).values
#     #     r_f = self.foreign_zero_curve.zero_rate(dates=expiry_datetimeindex,compounding_frequency=CompoundingFrequency.CONTINUOUS).values
#     #     F = interp_fx_forward_curve(self.fx_forward_curve, dates=expiry_datetimeindex, date_type='expiry_date', flat_extrapolation=True)
#     #
#     #     results = gk_price(
#     #         S0=self.fx_spot_rate,
#     #         tau=year_fraction(self.curve_date, expiry_datetimeindex, self.day_count_basis),
#     #         r_d=r_d,
#     #         r_f=r_f,
#     #         cp=cp,
#     #         K=K,
#     #         vol=vol,
#     #         F=F,
#     #         analytical_greeks_flag=analytical_greeks_flag,
#     #         intrinsic_time_split_flag=intrinsic_time_split_flag
#     #     )
#     #
#     #     results['market_data_inputs'] = pd.DataFrame({'vol': vol, 'r_d':r_d, 'r_f':r_f , 'F':F})
#     #
#     #     return results
#
#     #
#     # def simulate_fx_rate_path(self,
#     #                           date_grid=None,
#     #                           nb_simulations=None,
#     #                           flag_apply_antithetic_variates=None,
#     #                           method: str='geometric_brownian_motion'):
#     #
#     #     results = dict()
#     #
#     #     if date_grid is None:
#     #         date_grid = self.σ_pillar.index # create the date_grid based on the volatility input tenors
#     #
#     #     if method == 'geometric_brownian_motion':
#     #
#     #         if self.curve_date not in date_grid:
#     #             date_grid = pd.DatetimeIndex([self.curve_date]).append(date_grid)
#     #
#     #         date_grid = date_grid.unique().sort_values(ascending=True)
#     #         tau = year_fraction(self.curve_date, date_grid, self.day_count_basis).values # this should be based on expiry_dates
#     #         dt = tau[1:] - tau[:-1]
#     #         dt = np.insert(dt, 0, 0)
#     #
#     #         fx_forward_rates_t2 = interp_fx_forward_curve(self.fx_forward_curve, dates=date_grid, date_type='delivery_date').values
#     #         fx_forward_rates_t1 = np.zeros(shape=fx_forward_rates_t2.shape)
#     #         fx_forward_rates_t1[0] = self.fx_spot_rate
#     #         fx_forward_rates_t1[1:] = fx_forward_rates_t2[:-1].copy()
#     #         period_drift = np.log(fx_forward_rates_t2 / fx_forward_rates_t1) / dt
#     #         period_drift[0] = np.nan
#     #
#     #         # tk need to update this to do with delivery dates no index's
#     #         mask = np.logical_and.reduce([date_grid >= self.σ_pillar.index.min(), date_grid <= self.σ_pillar.index.max()])
#     #         σ_atm = np.full(date_grid.shape, np.nan)
#     #         if mask.any():
#     #             σ_atm[mask] = self.daily_volatility_smile['σ_atmΔneutral'][date_grid[mask]]
#     #
#     #         # Apply flat extrapolation for dates outside the volatility range
#     #         σ_atm[date_grid < self.σ_pillar.index.min()] = self.daily_volatility_smile['σ_atmΔneutral'][date_grid[mask]][0]
#     #         σ_atm[date_grid > self.σ_pillar.index.max()] = self.daily_volatility_smile['σ_atmΔneutral'][date_grid[mask]][-1]
#     #
#     #         t1 = tau[:-1]
#     #         t2 = tau[1:]
#     #         σ_atm_t1 = σ_atm[:-1].copy()
#     #         σ_atm_t2 = σ_atm[1:].copy()
#     #         σ_forward = self.forward_volatility(t1, σ_atm_t1, t2, σ_atm_t2)
#     #
#     #         # Setup data frame with the key market data inputs for easier review
#     #         df_gbm_monte_carlo_market_data_inputs = pd.DataFrame()
#     #         df_gbm_monte_carlo_market_data_inputs['dates'] = date_grid.values
#     #         df_gbm_monte_carlo_market_data_inputs['tau'] = tau
#     #         df_gbm_monte_carlo_market_data_inputs['dt'] = dt
#     #         df_gbm_monte_carlo_market_data_inputs['fx_forward_rates'] = fx_forward_rates_t2
#     #         df_gbm_monte_carlo_market_data_inputs['drift'] = period_drift
#     #         df_gbm_monte_carlo_market_data_inputs['atm_volatility'] = σ_atm
#     #         df_gbm_monte_carlo_market_data_inputs['forward_atm_volatility'] = np.insert(σ_forward, 0, np.nan)
#     #         results['gbm_monte_carlo_market_data_inputs'] = df_gbm_monte_carlo_market_data_inputs
#     #
#     #         rand_nbs = generate_rand_nbs(nb_steps=len(date_grid)-1,
#     #                                      nb_rand_vars=1,
#     #                                      nb_simulations=nb_simulations,
#     #                                      flag_apply_antithetic_variates=flag_apply_antithetic_variates)
#     #
#     #         fx_rate_simulation_paths = simulate_gbm_path(initial_px=self.fx_spot_rate,
#     #                                                      drift=period_drift[1:],
#     #                                                      forward_volatility=σ_forward,
#     #                                                      timestep_length=dt[1:],
#     #                                                      rand_nbs=rand_nbs)
#     #         results['fx_rate_simulation_paths'] = fx_rate_simulation_paths
#     #
#     #         return results
#     #
#     #
#     #     elif method == 'heston':
#     #
#     #         # Interpolate the volatility smile for the given expiry's
#     #         self.interp_σ_smile_from_pillar_points(expiry_dates=date_grid) # this function should be able to be called by delivery and expiry
#     #
#     #         # Date grid is defined as the delivery date.
#     #         # Need to update entire volatility surface with two indices, expiry and delivery date
#     #
#     #         # It doesn't really make sense to simulate a rate path with just a Heston model.
#     #         # The standard heston model, is best for pricing exotics / options where the only input is the termiminal fx rate
#     #         # Need the Heston model to be linked to the term structure - i.e. Local Stochastic volatility model
#     #
#     #         results = {}
#     #
#     #         for i,delivery_date in enumerate(date_grid):
#     #
#     #             print(i, delivery_date)
#     #
#     #             tau = year_fraction(self.curve_date,delivery_date, self.day_count_basis)
#     #             fx_forward_rate = interp_fx_forward_curve(self.fx_forward_curve, dates=pd.DatetimeIndex([delivery_date]), date_type='delivery_date').values
#     #             mu = np.log(fx_forward_rate/self.fx_spot_rate) / tau
#     #
#     #             var0 = self.daily_volatility_smile_func[delivery_date]['var0']
#     #             vv = self.daily_volatility_smile_func[delivery_date]['vv']
#     #             kappa = self.daily_volatility_smile_func[delivery_date]['kappa']
#     #             theta = self.daily_volatility_smile_func[delivery_date]['theta']
#     #             rho = self.daily_volatility_smile_func[delivery_date]['rho']
#     #             # lambda_ is not used in the simulation, used in the calibration only.
#     #             # lambda_ = self.daily_volatility_smile_func[delivery_date]['lambda_']
#     #
#     #             rand_nbs = generate_rand_nbs(
#     #                 nb_steps=100,
#     #                 nb_rand_vars=2,
#     #                 nb_simulations=10*1000,
#     #                 flag_apply_antithetic_variates=False)
#     #
#     #             sim_results = simulate_heston(
#     #                 S0=self.fx_spot_rate,
#     #                 mu=mu,
#     #                 var0=var0,
#     #                 vv=vv,
#     #                 kappa=kappa,
#     #                 theta=theta,
#     #                 rho=rho,
#     #                 tau=tau,
#     #                 rand_nbs=rand_nbs)
#     #
#     #             results[delivery_date] = sim_results
#     #
#     #         return results
#     #
#
#
#
#
#
#



# -*- coding: utf-8 -*-
import os
if __name__ == "__main__":
    os.chdir(os.environ.get('PROJECT_DIR_FRM'))
from dataclasses import dataclass, field
import numpy as np
import pandas as pd
import scipy
from frm.pricing_engine.black76_bachelier import black76_price, bachelier_price, black76_solve_implied_vol, bachelier_solve_implied_vol, normal_vol_to_black76_sln, black76_sln_to_normal_vol, black76_sln_to_normal_vol_analytical, normal_vol_atm_to_black76_sln_atm, VOL_SLN_BOUNDS, VOL_N_BOUNDS
from frm.pricing_engine.sabr import solve_alpha_from_sln_vol, calc_sln_vol_for_strike_from_sabr_params
from frm.utils import year_frac, clean_tenor, tenor_to_date_offset, convert_column_to_consistent_data_type, Schedule, get_schedule
from frm.enums import DayCountBasis, PeriodFreq, TermRate
from frm.term_structures.zero_curve import ZeroCurve
from typing import Optional, Union, List
import time
import numbers
from frm.term_structures.interest_rate_option_helpers import standardise_relative_quote_col_names, standardise_atmf_quote_col_names



@dataclass
class Optionlet:
    curve_date: pd.Timestamp
    quote_vol_sln: pd.DataFrame
    ln_shift: float
    zero_curve: ZeroCurve
    optionlet_freq: PeriodFreq
    settlement_delay: Optional[int]=None
    settlement_date: Optional[pd.Timestamp]=None
    busdaycal: Optional[np.busdaycalendar]=np.busdaycalendar()


    #### Attributes set in __post_init__
    day_count_basis: DayCountBasis=field(init=False)

    quote_columns: list = field(init=False)
    quote_strikes: pd.DataFrame = field(init=False)
    quote_cp: pd.DataFrame = field(init=False)
    quote_pxs: pd.DataFrame = field(init=False)
    quote_vol_n: pd.DataFrame = field(init=False)

    term_structure: pd.DataFrame = field(init=False)

    def __post_init__(self):

        t = time.time()
        self.day_count_basis = self.zero_curve.day_count_basis

        # Identical for Cap/Floor and IRS bootstrapping
        if self.settlement_date is None and self.settlement_delay is None:
            raise ValueError('Either settlement_date or settlement_delay must be provided.')
        elif self.settlement_date is None and self.settlement_delay is not None:
            self.settlement_date = np.busday_offset(self.curve_date.to_numpy().astype('datetime64[D]'),
                                                    offsets=self.settlement_delay,
                                                    roll='following',
                                                    busdaycal=self.busdaycal)

        # TODO This should be a setting - settlement_date + freq (as below) or effective at the settlement date.
        effective_date_np = (self.settlement_date + self.optionlet_freq.date_offset).to_numpy().astype('datetime64[D]')
        effective_date = np.busday_offset(effective_date_np, offsets=0, roll='following', busdaycal=self.busdaycal)
        self.quote_vol_sln['tenor'] = self.quote_vol_sln['tenor'].apply(clean_tenor)
        self.quote_vol_sln['settlement_date'] = self.settlement_date
        self.quote_vol_sln['effective_date'] = effective_date
        self.quote_vol_sln['last_optionlet_expiry_date'] = np.nan
        self.quote_vol_sln['termination_date'] = np.nan

        for i, row in self.quote_vol_sln.iterrows():
            date_offset = tenor_to_date_offset(row['tenor'])
            termination_date = self.settlement_date + date_offset
            last_optionlet_expiry_date = termination_date - self.optionlet_freq.date_offset
            last_optionlet_expiry_date_np = last_optionlet_expiry_date.to_numpy().astype('datetime64[D]')
            termination_date_date_np = termination_date.to_numpy().astype('datetime64[D]')
            self.quote_vol_sln.at[i, 'last_optionlet_expiry_date'] = np.busday_offset(
                last_optionlet_expiry_date_np, offsets=0,roll='following', busdaycal=self.busdaycal)
            self.quote_vol_sln.at[i, 'termination_date'] = np.busday_offset(
                termination_date_date_np, offsets=0,roll='following', busdaycal=self.busdaycal)

        self.quote_vol_sln['term_years'] = year_frac(
            self.quote_vol_sln['effective_date'], self.quote_vol_sln['termination_date'], self.day_count_basis)
        self.quote_vol_sln['last_optionlet_expiry_years'] = year_frac(
            self.curve_date, self.quote_vol_sln['last_optionlet_expiry_date'],self.day_count_basis)
        self.quote_vol_sln['F'] = np.nan

        self.quote_vol_sln = convert_column_to_consistent_data_type( self.quote_vol_sln)

        first_columns = ['tenor', 'settlement_date', 'effective_date', 'last_optionlet_expiry_date', 'termination_date',
                         'term_years', 'last_optionlet_expiry_years', 'F']
        column_order = first_columns + [col for col in  self.quote_vol_sln.columns if col not in first_columns]
        self.quote_vol_sln =  self.quote_vol_sln[column_order]
        self.quote_vol_sln.sort_values(by=['termination_date'], inplace=True)
        self.quote_vol_sln.reset_index(drop=True, inplace=True)

        # Two methods of specifying quotes
        # 1. Quotes relative to the atm forward rate (e.g. ATM, ATM+/-50bps, ATM+/-100bps...)
        # 2. Absolute quotes (e.g. 2.5%, 3.0%, 3.5%...)

        # Method 1
        col_name_update, quote_str_map = standardise_relative_quote_col_names(col_names=list(self.quote_vol_sln.columns))
        self.quote_vol_sln.rename(columns=col_name_update, inplace=True)

        self.quote_vol_sln.reset_index(drop=True, inplace=True)
        self.quote_cols = list(quote_str_map.keys())
    
        nb_quotes = len(self.quote_vol_sln) - 1
        self.term_structure = get_schedule(start_date=self.quote_vol_sln.loc[nb_quotes, 'effective_date'],
                                           end_date=self.quote_vol_sln.loc[nb_quotes, 'termination_date'],
                                           freq=self.optionlet_freq,
                                           cal=self.cal)
        self.term_structure.add_period_yearfrac(day_count_basis=self.day_count_basis)

        self.term_structure['discount_factors'] = self.zero_curve.get_discount_factors(dates=self.term_structure['payment_date'])
        self.term_structure['annuity_factor'] = self.term_structure['period_yearfrac'] * self.term_structure['discount_factors']
        self.term_structure['expiry_years'] = year_frac(self.curve_date, self.term_structure['period_start'], self.day_count_basis)
        self.term_structure['F'] = self.zero_curve.get_forward_rates(period_start=self.term_structure['period_start'],
                                                                     period_end=self.term_structure['period_end'],
                                                                     forward_rate_type=TermRate.SIMPLE)

        self.term_structure[['vol_n_atm', 'vol_sln_atm', 'alpha', 'beta', 'rho', 'volvol']] = np.nan

        self.term_structure.insert(loc=0, column='quote_nb', value=np.nan)
        for i, row in self.term_structure.iterrows():
            mask = row['period_end'] <= self.quote_vol_sln['termination_date']
            if mask.any():
                last_valid_index = mask.idxmax()
            else:
                last_valid_index = 0
            self.term_structure.at[i, 'quote_nb'] = last_valid_index

        # Calculate the forward rate (pre lognormal shift) for the term structure
        for i, row in self.quote_vol_sln.iterrows():
            mask = (self.term_structure['period_end'] <= self.quote_vol_sln.loc[i, 'termination_date'])
            self.quote_vol_sln.loc[i, 'F'] = \
                (self.term_structure.loc[mask, 'F'] * self.term_structure.loc[mask, 'annuity_factor']).sum() \
                / self.term_structure.loc[mask, 'annuity_factor'].sum()

        # Setup a strikes dataframe for the cap/floor quotes
        self.quote_strikes = self.quote_vol_sln.copy()
        self.quote_strikes[self.quote_cols] = np.nan
        for col_name in self.quote_strikes[self.quote_cols].columns:
            self.quote_strikes[col_name] = self.quote_strikes['F'] + quote_str_map[col_name]

        # Setup a call/put flag dataframe for the cap/floor quotes
        self.quote_cp = self.quote_vol_sln.copy()
        self.quote_cp[self.quote_cols] = np.nan
        self.quote_cp[self.quote_cols] = \
            np.where(self.quote_strikes[self.quote_cols].values > self.quote_cp['F'].values[:, None], 1, -1)

        # Price the cap/floors per the lognormal volatility quotes
        self.quote_pxs = pd.DataFrame(
            data=np.full(shape=(len(self.quote_vol_sln), len(self.quote_cols)), fill_value=np.nan),
            columns=self.quote_cols)

        for quote_nb, _ in self.quote_vol_sln.iterrows():
            mask = self.term_structure['quote_nb'] <= quote_nb
            optionlet_pxs = np.full(shape=(len(self.term_structure[mask]), len(self.quote_cols)), fill_value=np.nan)
            cp = self.quote_cp.loc[quote_nb, self.quote_cols].astype('float64').values
            K = self.quote_strikes.loc[quote_nb, self.quote_cols].astype('float64').values
            vol_sln = self.quote_vol_sln.loc[quote_nb, self.quote_cols].astype('float64').values
            ln_shift = self.ln_shift

            for i, row in self.term_structure[mask].iterrows():
                optionlet_pxs[i, :] = black76_price(
                    F=row['F'],
                    tau=row['expiry_years'],
                    cp=cp,
                    K=K,
                    vol_sln=vol_sln,
                    ln_shift=ln_shift)['price'] * row['annuity_factor']
            try:
                self.quote_pxs.loc[quote_nb, self.quote_cols] = optionlet_pxs.sum(axis=0)
            except:
                print(quote_nb, self.quote_cols, optionlet_pxs)
                raise

        t1 = time.time()

        # Solve the equivalent normal volatilities for to the lognormal cap/floor volatility quotes.
        # Can be parallelized as each solve is independent
        self.quote_vol_n = self.quote_vol_sln.copy()
        self.quote_vol_n[self.quote_cols] = np.nan

        def obj_func_relative_px_error(vol_n):
            """Helper function for solving the equivalent normal (bachelier) volatility for a cap/floor quote"""
            bachelier_optionlet_pxs = bachelier_price(F=F, tau=tau, cp=cp, K=K, vol_n=vol_n.item(), annuity_factor=annuity_factor)['price']
            # Return the square of the relative error (to the price, as we want invariance to the price) to be minimised.
            # Squared error ensures differentiability which is critical for gradient-based optimisation methods.
            return ((target - bachelier_optionlet_pxs.sum(axis=0)) / target) ** 2

        # Want prices to be within 0.001% - i.e. if price is 100,000, acceptable range is (99999, 100001)
        # Hence set obj function tol 'ftol' to (0.001%)**2 = 1e-10 (due to the squaring of the relative error in the obj function)
        # We set gtol to zero, so that the optimisation is terminated based on ftol.
        obj_func_tol = 1e-5 ** 2  # 0.001%^2
        options = {'ftol': obj_func_tol, 'gtol': 0}

        for quote_nb, row in self.quote_vol_sln.iterrows():
            mask = self.term_structure['quote_nb'] <= quote_nb
            F = self.term_structure.loc[mask, 'F'].values
            tau = self.term_structure.loc[mask, 'expiry_years'].values
            annuity_factor = self.term_structure.loc[mask, 'annuity_factor'].values

            for quote_col in self.quote_cols:
                cp = self.quote_cp.loc[quote_nb, quote_col]
                K = self.quote_strikes.loc[quote_nb, quote_col]
                target = self.quote_pxs.loc[quote_nb, quote_col]

                x0 = np.atleast_1d(black76_sln_to_normal_vol_analytical(
                        F=row['F'],
                        tau=row['last_optionlet_expiry_years'],
                        K=K,
                        vol_sln=self.quote_vol_sln.loc[quote_nb, quote_col],
                        ln_shift=self.ln_shift))

                res = scipy.optimize.minimize(
                    fun=obj_func_relative_px_error,
                    x0=x0,
                    bounds=[(x0*0.8, x0*1.25)],
                    method='L-BFGS-B',
                    options=options)

                # 2nd condition required as we have overridden options to terminate based on ftol
                if res.success or abs(res.fun) < obj_func_tol:
                    self.quote_vol_n.loc[quote_nb, quote_col] = res.x.item()
                else:
                    print(res)
                    raise ValueError('Optimisation of lognormal volatility quotes, normal volatility equivalent, failed to converge')

        print('time to solve quotes equivalent normal vol:', time.time() - t1)

        solve_vol_n_t = []
        solve_sabr_t = []

        # Bootstrap the optionlet volatilities and fit the SABR smiles
        for quote_nb in range(len(self.quote_vol_sln)):

            if 'beta' in self.quote_vol_sln.columns and not pd.isnull(self.quote_vol_sln.loc[quote_nb, 'beta']):
                beta_overide = self.quote_vol_sln.loc[quote_nb, 'beta']
            else:
                beta_overide = None
            t1 = time.time()
            self._solve_vol_n_atm(quote_nb=quote_nb)
            t2 = time.time()
            self._solve_sabr_params(quote_nb=quote_nb, beta_overide=beta_overide)
            t3 = time.time()

            solve_vol_n_t.append(t2-t1)
            solve_sabr_t.append(t3-t2)

        print('time to solve_vol_n_atm:', np.sum(solve_vol_n_t))
        print('time to solve_sabr_params:', np.sum(solve_sabr_t))


    def _solve_vol_n_atm(self, quote_nb):
        """
        Function for bootstrapping the normal at-the-money forward (ATMF) volatilities for the optionlets term structure.

        When bootstrapping an optionlet term structure, one methodology decision is whether to
        (a) calculate alpha from the other SABR parameters, or
        (b) include alpha in the calibration of the SABR parameters.

        This function is a helper function for the optionlet bootstrapping process under decision (a).
        Method (a) requires the optionlet ATMF volatilities to be bootstrapped first,
        as the optionlet ATMF is an input to the analytical formula for alpha.

        Assumptions:
        For term of the first quote, the ATMF normal volatility of the component optionlets is set to be equal to the quote.
        For intra-quote optionlets, the ATMF normal volatility is linearly interpolated between pillar optionlets.
        """

        mask_prior = self.term_structure['quote_nb'] < quote_nb
        mask_current = self.term_structure['quote_nb'] == quote_nb
        prior_pillar_idx = pd.Series(mask_prior)[::-1].idxmax()
        current_pillar_idx = pd.Series(mask_current)[::-1].idxmax()

        if quote_nb == 0:
            # Set the ATMF normal volatility of the component optionlets of the 1st ATMF quote to be equal to the quote.
            self.term_structure.loc[mask_current, 'vol_n_atm'] = self.quote_vol_n.loc[quote_nb, 'atmf']
        elif quote_nb > 0:
            # For 2nd, 3rd... ATMF quotes, an optimisation is required to solve the optionlets normal ATMF volatilities.
            #
            # We must solve (b) to equal (a) where:
            # (a) is the optionlets priced with the flat/scalar ATMF normal volatilities quote, and
            # (b) is the optionlets priced with a term structure of ATMF normal volatilities, interpolated of the SABR smile.
            #
            # (b) is composed of:
            # (bi) pricing the optionlets with the SABR smile already solved.
            # (bii) numerically solve the normal ATMF volatility at the quotes terminal 'pillar' optionlet in order for (bi) + (bii) to match (a)

            # (bi): Price the optionlets where with the SABR smile already solved.

            optionlets = self.term_structure.loc[mask_prior, :].copy()

            vol_sln = calc_sln_vol_for_strike_from_sabr_params(
                    tau=optionlets['expiry_years'].values,
                    F=optionlets['F'].values,
                    alpha=optionlets['alpha'].values,
                    beta=optionlets['beta'].values,
                    rho=optionlets['rho'].values,
                    volvol=optionlets['volvol'].values,
                    K=self.quote_strikes.loc[quote_nb, 'atmf'],
                    ln_shift=self.ln_shift)

            optionlet_px = black76_price(
                F=optionlets['F'].values,
                tau=optionlets['expiry_years'].values,
                cp=self.quote_cp.loc[quote_nb, 'atmf'],
                K=self.quote_strikes.loc[quote_nb, 'atmf'],
                vol_sln=vol_sln,
                ln_shift=self.ln_shift)['price'] * optionlets['annuity_factor'].values

            N = len(optionlets) - 1
            normal_vol_prior_pillar = black76_sln_to_normal_vol(
                F=optionlets.loc[N,'F'],
                tau=optionlets.loc[N,'expiry_years'],
                K=self.quote_strikes.loc[quote_nb, 'atmf'],
                vol_sln=vol_sln[-1],
                ln_shift=self.ln_shift)

            # The target of the numerical solve is (a) - (bi)
            target = self.quote_pxs.loc[quote_nb, 'atmf'] - sum(optionlet_px)

            # Index and define the parameters of the numerical solve to variables outside the function for speed.
            X = [self.term_structure.loc[prior_pillar_idx, 'expiry_years'],
                 self.term_structure.loc[current_pillar_idx, 'expiry_years']]
            x = self.term_structure.loc[mask_current, 'expiry_years'].values
            F = self.term_structure.loc[mask_current, 'F'].values
            tau = self.term_structure.loc[mask_current, 'expiry_years'].values
            cp = self.quote_cp.loc[quote_nb, 'atmf']
            K = self.quote_strikes.loc[quote_nb, 'atmf']
            annuity_factor = self.term_structure.loc[mask_current, 'annuity_factor'].values

            def vol_n_atm_obj_func(param):
                """"Error function for the optimisation of the normal at-the-money volatility for pillar optionlet."""
                try:
                    Y = [normal_vol_prior_pillar, param.item()]
                    vol_n = np.interp(x, X, Y)
                except ValueError:
                    print(f'X: {X}, Y: {Y}, x: {x}')
                    raise
                optionlet_pxs = bachelier_price(F=F, tau=tau, cp=cp, K=K, vol_n=vol_n)['price'] * annuity_factor
                # Return the square of the relative error (to the price, as we want invariance to the price) to be minimised.
                # Squared error ensures differentiability which is critical for gradient-based optimisation methods.
                return ((target - optionlet_pxs.sum(axis=0)) / target)**2

            # (bii): Numerically solve the normal ATMF volatility for the terminal 'pillar' optionlet.
            # Want prices to be within 0.001% - i.e. if price is 100,000, acceptable range is (99999, 100001)
            # Hence set obj function tol 'ftol' to (0.001%)**2 = 1e-10 (due to the squaring of the relative error in the obj function)
            # We set gtol to zero, so that the optimisation is terminated based on ftol.
            obj_func_tol = 1e-5**2 # 0.001%^2
            options = {'ftol': obj_func_tol, 'gtol': 0}
            res = scipy.optimize.minimize(
                fun=vol_n_atm_obj_func,
                x0=np.atleast_1d(normal_vol_prior_pillar),
                bounds=[VOL_N_BOUNDS],
                method='L-BFGS-B',
                options=options)

            # 2nd condition required as we have overridden options to terminate based on ftol
            if res.success or abs(res.fun) < obj_func_tol:
                vol_n_atm_pillar = res.x.item()
            else:
                print(res)
                raise ValueError('Optimisation of pillar optionlet normal volatility failed to converge')

            # For the intra-pillar optionlets, interpolate the normal ATMF volatility between the pillar points.
            # Linear interpolation is used, this is an assumption. Other interpolation methods could be used.
            X = [self.term_structure.loc[prior_pillar_idx, 'expiry_years'],
                 self.term_structure.loc[current_pillar_idx, 'expiry_years']]
            Y = [self.term_structure.loc[prior_pillar_idx, 'vol_n_atm'], vol_n_atm_pillar]
            x = self.term_structure.loc[mask_current, 'expiry_years']
            self.term_structure.loc[mask_current, 'vol_n_atm'] = np.interp(x, X, Y)

        # Solve the equivalent log-normal volatilities for each optionlet.
        for i, row in self.term_structure.loc[mask_current,:].iterrows():
            self.term_structure.loc[i, 'vol_sln_atm'] = normal_vol_to_black76_sln(
                F=row['F'],
                tau=row['expiry_years'],
                K=row['F'],
                vol_n=row['vol_n_atm'],
                ln_shift=self.ln_shift)


    def _solve_sabr_params(self, quote_nb, beta_overide=None):
        """
        Function for calibrating the SABR parameters for the optionlet term structure based on cap/floor lognormal volatility quotes.
        """

        def sse_sabr_for_pillar_optionlet(self, param):
            """
            Helper function producing the sum of squared errors between
            (a) cap/floor prices based on the quoted flat/scalar lognormal volatility quotes
            (b) repricing where each component optionlet uses the volatility interpolated of its SABR smiles.
            """

            if beta_overide is None:
                beta_pillar, rho_pillar, volvol_pillar = param
            else:
                rho_pillar, volvol_pillar = param
                beta_pillar = beta_overide

            optionlet_pxs = np.full(shape=(len(self.term_structure[mask_current]), len(self.quote_cols)), fill_value=np.nan)
            K = self.quote_strikes.loc[quote_nb, self.quote_cols].astype('float64').values

            if quote_nb == 0:
                # For term of the 1st quote, solve for a flat beta, rho & volvol.
                N = len(self.term_structure[mask_current])
                beta = [beta_pillar] * N
                rho = [rho_pillar] * N
                volvol = [volvol_pillar] * N
            else:
                # The solved SABR parameters define the terminal/'pillar' optionlet for the current quote.
                # For the intra-quote optionlets, interpolate the SABR parameters between the prior and current SABR parameters.
                # Linear interpolation is applied, this is an assumption. Other interpolation methods could be used.
                X = [self.term_structure.loc[prior_pillar_idx, 'expiry_years'],
                     self.term_structure.loc[current_pillar_idx, 'expiry_years']]
                x = self.term_structure.loc[mask_current, 'expiry_years'].values

                beta = np.interp(x, X, [self.term_structure.loc[prior_pillar_idx, 'beta'], beta_pillar])
                rho = np.interp(x, X, [self.term_structure.loc[prior_pillar_idx, 'rho'], rho_pillar])
                volvol = np.interp(x, X,[self.term_structure.loc[prior_pillar_idx, 'volvol'], volvol_pillar])

            for idx, (i, row) in enumerate(self.term_structure.loc[mask_current, :].iterrows()):
                alpha = solve_alpha_from_sln_vol(tau=row['expiry_years'], F=row['F'], beta=beta[idx], rho=rho[idx], volvol=volvol[idx], vol_sln_atm=row['vol_sln_atm'],ln_shift=self.ln_shift)
                vols_sln = calc_sln_vol_for_strike_from_sabr_params(tau=row['expiry_years'], F=row['F'], alpha=alpha, beta=beta[idx], rho=rho[idx], volvol=volvol[idx], K=K, ln_shift=self.ln_shift)

                optionlet_pxs[idx, :] = black76_price(
                    F=row['F'],
                    tau=row['expiry_years'],
                    cp=self.quote_cp.loc[quote_nb, self.quote_cols].astype('float64').values,
                    K=K,
                    vol_sln=vols_sln,
                    ln_shift=self.ln_shift)['price'] * row['annuity_factor']

            relative_error_of_pxs = (target - optionlet_pxs.sum(axis=0)) / target
            return sum((relative_error_of_pxs ** 2) / len(relative_error_of_pxs))

        mask_prior = self.term_structure['quote_nb'] < quote_nb
        mask_current = self.term_structure['quote_nb'] == quote_nb
        prior_pillar_idx = pd.Series(mask_prior)[::-1].idxmax()
        current_pillar_idx = pd.Series(mask_current)[::-1].idxmax()

        # We solve the SABR smile for the pillar optionlet in order to match
        # (a) the pricing using the scalar (flat) volatility quote.
        # (b) the cap/floor quote pricing using a volatility optionlet term structure with SABR smile.

        # (b) is composed of:
        # (bi) pricing the optionlets where the SABR smile has already be solved in prior iterations.
        # (bii) numerically solve the SABR parameters at the quotes terminal 'pillar' optionlet in order for (bi) + (bii) to match (a)

        # (bi): Price the optionlets where with the SABR smile already solved.
        if quote_nb == 0:
            # For the 1st pillar point (bi)=0 so we just solve (bii) to equal (a).
            target = self.quote_pxs.loc[quote_nb, self.quote_cols]
        else:
            # For the 2nd, 3rd... pillar points, we first must solve (bi).
            prior_optionlet_pxs = np.full(shape=(len(self.term_structure[mask_prior]), len(self.quote_cols)), fill_value=np.nan)
            for i, row in self.term_structure.loc[mask_prior, :].iterrows():
                vol_sln = calc_sln_vol_for_strike_from_sabr_params(
                    tau=row['expiry_years'],
                    F=row['F'],
                    alpha=row['alpha'],
                    beta=row['beta'],
                    rho=row['rho'],
                    volvol=row['volvol'],
                    K=self.quote_strikes.loc[quote_nb, self.quote_cols].astype('float64').values,
                    ln_shift=self.ln_shift)

                prior_optionlet_pxs[i, :] = black76_price(
                    F=row['F'],
                    tau=row['expiry_years'],
                    cp=self.quote_cp.loc[quote_nb, self.quote_cols].astype('float64').values,
                    K=self.quote_strikes.loc[quote_nb, self.quote_cols].astype('float64').values,
                    vol_sln=vol_sln,
                    ln_shift=self.ln_shift)['price'] * row['annuity_factor']

            target = self.quote_pxs.loc[quote_nb, self.quote_cols] - prior_optionlet_pxs.sum(axis=0)

        # alpha has a valid range of 0≤α≤∞, though it is analytically calculated from the other SABR parameters.
        # beta has a valid range of 0≤β≤1 and is either solved or specified.
        # rho has a valid range of -1≤ρ≤1
        # volvol has a valid range of 0<v≤∞
        x0 = np.array([0.00, 0.10]) if beta_overide is not None else np.array([0.0, 0.0, 0.1])
        bounds = [(-1.0, 1.0), (0.0001, None)] if beta_overide is not None else [(-1.0, 1.0), (-1.0, 1.0),(0.0001, None)]
        res = scipy.optimize.minimize(fun=lambda param: sse_sabr_for_pillar_optionlet(self=self, param=param),
                                      x0=x0,
                                      bounds=bounds,
                                      tol=1e-5)
        if res.success:
            beta_pillar, rho_pillar, volvol_pillar = (beta_overide, *res.x) if beta_overide is not None else res.x
        else:
            print(res)
            raise ValueError('Optimisation of SABR parameters failed to converge')

        if quote_nb == 0:
            # For term of the 1st cap/floor quote, we assume a flat beta, rho & volvol over the term.
            N = len(self.term_structure[mask_current])
            beta = [beta_pillar] * N
            rho = [rho_pillar] * N
            volvol = [volvol_pillar] * N
        else:
            # For the 2nd, 3rd... cap/floor quotes, we interpolate the SABR parameters of intra-pillar between the prior and current pillar points.
            X = [self.term_structure.loc[prior_pillar_idx, 'expiry_years'],
                 self.term_structure.loc[current_pillar_idx, 'expiry_years']]
            x = self.term_structure.loc[mask_current, 'expiry_years'].values
            beta = np.interp(x, X, [self.term_structure.loc[prior_pillar_idx, 'beta'], beta_pillar])
            rho = np.interp(x, X, [self.term_structure.loc[prior_pillar_idx, 'rho'], rho_pillar])
            volvol = np.interp(x, X, [self.term_structure.loc[prior_pillar_idx, 'volvol'], volvol_pillar])

        # Store the optionlets SABR params in 'term_structure'
        for idx, (i, row) in enumerate(self.term_structure.loc[mask_current, :].iterrows()):
            # Alpha is analytically solved from the other SABR parameters, the ATMF optionlet volatility and the forward rate.
            alpha = solve_alpha_from_sln_vol(tau=row['expiry_years'], F=row['F'], beta=beta[idx], rho=rho[idx], volvol=volvol[idx], vol_sln_atm=row['vol_sln_atm'], ln_shift=self.ln_shift)
            self.term_structure.loc[i, 'alpha'] = alpha
            self.term_structure.loc[i, 'beta'] = beta[idx]
            self.term_structure.loc[i, 'rho'] = rho[idx]
            self.term_structure.loc[i, 'volvol'] = volvol[idx]


    def interpolate_sabr_term_structure(self, schedule_df: pd.DataFrame) -> pd.DataFrame:
        """ Interpolates the SABR term structure for a schedule."""
        #TODO - Add a check to ensure the effective date is within the term structure
        # Currently we are using the objects daycountbasis, calendar etc.
        #TODO - consider how to support stubs, non-default day count basis, payment delays, etc.
        # TODO - how to consider rolled effective dates be rolled if weekends or holidays

        F = self.zero_curve.get_forward_rates(period_start=schedule_df['period_start'],
                                              period_end=schedule_df['period_end'],
                                              forward_rate_type=TermRate.SIMPLE)

        # Selecting the specific columns for linear interpolation
        cols_to_interpolate = ['beta', 'rho', 'volvol', 'vol_n_atm']

        # Linearly interpolate SABR parameters for each date in 'period_start'
        interp_sabr_params = (
            pd.concat([self.term_structure, pd.DataFrame({'period_start': schedule_df['period_start']})])
            .set_index('period_start')
            .sort_index()[cols_to_interpolate]
            .interpolate(method='index')
        ).loc[schedule_df['period_start']]

        # Remove any duplicate (occurs if duplicate across term structure and effective_dates)
        interp_sabr_params = interp_sabr_params[~interp_sabr_params.index.duplicated(keep='first')]

        interp_sabr_params.loc[:,'expiry_years'] = year_frac(self.curve_date, schedule_df['period_start'], self.day_count_basis)
        interp_sabr_params.loc[:,'F'] = F
        interp_sabr_params.loc[:,['vol_sln_atm', 'alpha']] = np.nan

        interp_sabr_params.loc[:,'vol_sln_atm'] = normal_vol_atm_to_black76_sln_atm(
            F=interp_sabr_params['F'].values,
            tau=interp_sabr_params['expiry_years'].values,
            vol_n_atm=interp_sabr_params['vol_n_atm'].values,
            ln_shift=self.ln_shift)

        for i, (period_start, row) in enumerate(interp_sabr_params.iterrows()):
            interp_sabr_params.loc[period_start, 'alpha'] = solve_alpha_from_sln_vol(
                tau=row['expiry_years'], F=row['F'], beta=row['beta'], rho=row['rho'], volvol=row['volvol'],
                vol_sln_atm=row['vol_sln_atm'], ln_shift=self.ln_shift)

        column_order = [col for col in self.term_structure.columns if col in interp_sabr_params.columns]
        interp_sabr_params = interp_sabr_params[column_order]

        # Reindex to the dates & order in schedule_df
        result = pd.concat([interp_sabr_params.loc[row['period_start']] for _, row in schedule_df.iterrows()], axis=1).T
        result.reset_index(inplace=True, drop=True)

        # Horizontal concatenation of the schedule_df and the interpolated SABR parameters
        result = pd.concat([schedule_df.reset_index(drop=True), result], axis=1)

        return result


    def _validate_strike_and_option_type(self, K, cp, schedule_length):
        # Ensure that 'K' is either a scalar or an array/list of the same length as 'effective_dates'
        if isinstance(K, (list, np.ndarray)):
            assert len(K) == schedule_length, \
                "'K' must have the same length as 'effective_dates' if it's a list or array."
        else:
            assert isinstance(K, numbers.Real), "'K' must be a numeric scalar."

        # Ensure that 'cp' is either a scalar or an array/list of the same length as 'effective_dates' and that its values are -1 or 1
        if isinstance(cp, (list, np.ndarray)):
            assert len(cp) == schedule_length, \
                "'cp' must have the same length as 'effective_dates' if it's a list or array."
            assert all(val in [-1, 1] for val in cp), "'cp' must only contain values of -1 or 1."
        else:
            assert isinstance(cp, int) and cp in [-1, 1], "'cp' must be a scalar and must be either -1 or 1."

        # Ensure both 'K' and 'cp' are either both scalars or both arrays/lists of the same length
        assert isinstance(K, (list, np.ndarray)) == isinstance(cp, (list, np.ndarray)), \
            "'K' and 'cp' must both be scalars or both arrays/lists of the same length."


    def _price_optionlets(self,
                          schedule: Schedule,
                          K: Union[float, List[float], np.ndarray],
                          cp: Union[int, List[int], np.ndarray],
                          day_count_basis: DayCountBasis,
                          vol_sln_override: Union[float, np.ndarray] = None,
                          vol_n_override: Union[float, np.ndarray] = None,
                          use_term_structure: bool = False,
                          solve_equivalent_flat_vol=False) -> pd.DataFrame:

        schedule.add_period_yearfrac(day_count_basis=day_count_basis)
        schedule_length = len(schedule.df)

        # Validate inputs
        self._validate_strike_and_option_type(K, cp, schedule_length)

        detail_df = schedule.df.copy()

        detail_df = self.interpolate_sabr_term_structure(schedule_df=detail_df)

        # Compute volatilities
        if use_term_structure:
            detail_df['vol_sln'] = calc_sln_vol_for_strike_from_sabr_params(
                tau=detail_df['expiry_years'].values,
                F=detail_df['F'].values,
                alpha=detail_df['alpha'].values,
                beta=detail_df['beta'].values,
                rho=detail_df['rho'].values,
                volvol=detail_df['volvol'].values,
                K=K,
                ln_shift=self.ln_shift
            )
        elif vol_sln_override is not None:
            if isinstance(vol_sln_override, np.ndarray) and vol_sln_override.size == 1:
                vol_sln_override = vol_sln_override.item()
            detail_df['vol_sln'] = vol_sln_override
        elif vol_n_override is not None:
            if isinstance(vol_n_override, np.ndarray) and vol_n_override.size == 1:
                vol_n_override = vol_n_override.item()
            detail_df['vol_n'] = vol_n_override

        detail_df['K'] = K
        detail_df['cp'] = cp
        detail_df['discount_factor'] = self.zero_curve.get_discount_factors(dates=detail_df['payment_date'])
        detail_df['annuity_factor'] = detail_df['period_yearfrac'] * detail_df['discount_factor']

        # Choose the appropriate pricing model
        if vol_n_override is not None:
            # Use Bachelier model
            optionlet_pricing = bachelier_price(
                F=detail_df['F'].values,
                tau=detail_df['expiry_years'].values,
                cp=detail_df['cp'].values,
                K=detail_df['K'].values,
                vol_n=detail_df['vol_n'].values,
                annuity_factor=detail_df['annuity_factor'].values,
                intrinsic_time_split=True,
                analytical_greeks=True)
        else:
            # Use Black76 model
            optionlet_pricing = black76_price(
                F=detail_df['F'].values,
                tau=detail_df['expiry_years'].values,
                cp=detail_df['cp'].values,
                K=detail_df['K'].values,
                vol_sln=detail_df['vol_sln'].values,
                ln_shift=self.ln_shift,
                annuity_factor=detail_df['annuity_factor'].values,
                intrinsic_time_split=True,
                analytical_greeks=True)

        # Assign results
        detail_df['forward_price'] = optionlet_pricing['price'] / detail_df['annuity_factor']
        detail_df['price'] = optionlet_pricing['price']
        detail_df['intrinsic'] = optionlet_pricing['intrinsic']
        detail_df['time'] = optionlet_pricing['time']
        detail_df['delta'] = optionlet_pricing['analytical_greeks']['delta']
        detail_df['vega'] = optionlet_pricing['analytical_greeks']['vega']

        if use_term_structure and solve_equivalent_flat_vol:
            target = detail_df['price'].sum()

            # TODO for extremely out the money strikes, with short tenors, this solve does not work.
            # TODO for detailed output, include the optionlet specific vol_n that matches the vol_sln.
            #  This also allows for including the normal vega at the optionlet detail.

            vol_sln_flat = black76_solve_implied_vol(
                F=detail_df['F'].values,
                tau=detail_df['expiry_years'].values,
                cp=detail_df['cp'].values,
                K=detail_df['K'].values,
                ln_shift=self.ln_shift,
                X=target,
                vol_sln_guess=detail_df['vol_sln'].mean(),
                annuity_factor=detail_df['annuity_factor'].values)

            vol_n_flat = bachelier_solve_implied_vol(
                F=detail_df['F'].values,
                tau=detail_df['expiry_years'].values,
                cp=detail_df['cp'].values,
                K=detail_df['K'].values,
                X=target,
                vol_n_guess=detail_df['vol_n_atm'].mean(),
                annuity_factor=detail_df['annuity_factor'].values)

            return detail_df, vol_sln_flat, vol_n_flat
        else:
            return detail_df

    def price_optionlets_per_term_structure(self,
                                            schedule: Schedule,
                                            K: Union[float, List[float], np.ndarray],
                                            cp: Union[int, List[int], np.ndarray],
                                            day_count_basis: DayCountBasis,
                                            solve_equivalent_flat_vol=False) -> pd.DataFrame:
        return self._price_optionlets(
            schedule=schedule,
            K=K,
            cp=cp,
            day_count_basis=day_count_basis,
            use_term_structure=True,
            solve_equivalent_flat_vol=solve_equivalent_flat_vol
        )

    def price_optionlets_per_flat_vol_sln(self,
                                          schedule: Schedule,
                                          K: Union[float, List[float], np.ndarray],
                                          cp: Union[int, List[int], np.ndarray],
                                          day_count_basis: DayCountBasis,
                                          vol_sln_override: Union[float, np.ndarray]) -> pd.DataFrame:
        return self._price_optionlets(
            schedule=schedule,
            K=K,
            cp=cp,
            day_count_basis=day_count_basis,
            vol_sln_override=vol_sln_override
        )

    def price_optionlets_per_flat_vol_n(self,
                                        schedule: Schedule,
                                        K: Union[float, List[float], np.ndarray],
                                        cp: Union[int, List[int], np.ndarray],
                                        day_count_basis: DayCountBasis,
                                        vol_n_override: Union[float, np.ndarray]) -> pd.DataFrame:
        return self._price_optionlets(
            schedule=schedule,
            K=K,
            cp=cp,
            day_count_basis=day_count_basis,
            vol_n_override=vol_n_override
        )


























# -*- coding: utf-8 -*-
import os
import numpy as np
import pandas as pd
from scipy.stats import norm
from scipy.optimize import root_scalar, newton

if __name__ == "__main__":
    os.chdir(os.environ.get('PROJECT_DIR_FRM')) 

def garman_kohlhagen_price(
        S0: [float, np.ndarray],
        tau: [float, np.ndarray],
        r_d: [float, np.ndarray],
        r_f: [float, np.ndarray],
        cp: [int, np.ndarray],
        K: [float, np.ndarray],
        vol: [float, np.ndarray],
        F: [float, np.ndarray] = None,
        analytical_greeks: bool=False,
        numerical_greeks: bool=False,
        intrinsic_time_split: bool=False
        ) -> dict:
    """
    Garman-Kohlhagen European FX option pricing formula for:
    - option total, intrinsic, and time value (in the domestic currency per 1 unit of foreign currency notional).
    - analytical and numerical greeks (normalized to shifts applied per market convention).

    Parameters
    ----------
    S0 : float
        FX spot rate (specified in # of units of domestic currency per 1 unit of foreign currency).
    tau : float
        Time to expiry (in years).
    r_d : float
        Domestic risk-free interest rate (annualized continuously compounded).
    r_f : float
        Foreign risk-free interest rate (annualized continuously compounded).
    cp : int
        Option type: 1 for call option, -1 for put option.
    K : float
        Strike price (in units of domestic currency per 1 unit of foreign currency).
    vol : float
        Volatility (annualized).
    F : float, optional
        Market forward rate. If None, it will be calculated using interest rate parity (default is None).
    analytical_greeks : bool, optional
        If True, analytical greeks will be calculated and returned (default is False).
    numerical_greeks : bool, optional
        If True, numerical greeks will be calculated and returned using finite differences (default is False).
    intrinsic_time_split : bool, optional
        If True, splits option value into intrinsic and time value components (default is False).

    Returns
    -------
    results : dict
        Dictionary containing the option price and, if requested, analytical or numerical greeks.
        Prices are in units of domestic currency per 1 unit of foreign currency.
        - 'price' : Option unit price
        - 'intrinsic' : Intrinsic unit price (if `intrinsic_time_split_flag` is True).
        - 'time' : Time unit price (if `intrinsic_time_split_flag` is True).
        - 'analytical_greeks' : DataFrame with analytical greeks (if `analytical_greeks_flag` is True).
        - 'numerical_greeks' : DataFrame with numerical greeks (if `numerical_greeks_flag` is True).
    Notes
    -----
    1. The option is priced under the assumption of continuous interest rate compounding.
    2. If the `F` parameter is provided, the domestic risk-free rate is adjusted to match the forward rate.
    3. The analytical greeks calculated are: delta, vega, gamma, theta, and rho.
    4. Numerical greeks are calculated using finite differences with small shifts (e.g., 1% for delta and vega, 1 day for theta).
    5. If tau equals 0, time value is set to zero. 
    """


    # Convert to arrays. Function is vectorised.
    S0, tau, r_d, r_f, cp, K, σ = map(lambda x: np.atleast_1d(x).astype(float), (S0, tau, r_d, r_f, cp, K, vol))

    # Value bounds checks. No >0 check for σ, as when doing numerical solving, need to allow for -ve σ
    assert (S0 > 0.0).all(), S0
    assert (tau >= 0.0).all(), tau 
    assert np.all(np.isin(cp, [1, -1])), cp
    assert cp.shape == σ.shape
    assert cp.shape == K.shape
    
    if F is not None: 
        # Use market forward rate and imply the currency basis-adjusted domestic interest rate
        F = np.atleast_1d(F).astype(float)
        assert (F > 0.0).all()
        r_d_basis_adj = np.log(F / S0) / tau + r_f # from F = S0 * exp((r_d - r_f) * tau)
        r = r_d_basis_adj
        q = r_f
    else:
        r = r_d
        q = r_f

    μ = r - q
    d1 = (np.log(S0 / K) + (μ + 0.5 * σ**2) * tau) / (σ * np.sqrt(tau))
    d2 = d1 - σ * np.sqrt(tau)   
    X = cp * (S0 * np.exp(-q * tau) * norm.cdf(cp * d1) - K * np.exp(-r * tau) * norm.cdf(cp * d2))

    X[tau==0] = np.maximum(0, cp * (S0 - K))[tau==0] # If time to maturity is 0.0, set to intrinsic value
    results = {'price': X}

    if intrinsic_time_split:
        intrinsic = np.full_like(X, np.nan)
        intrinsic[tau>0] = np.maximum(0, cp * (S0 * np.exp(-q * tau) - K * np.exp(-r * tau)))[tau>0]
        intrinsic[tau==0] = X[tau==0]
        results['intrinsic'] = intrinsic
        results['time'] = X - intrinsic

    if analytical_greeks:
        results['analytical_greeks'] = pd.DataFrame()

        # These formulae produce the delta applicable to the domestic currency notional.
        # To get the Δ % applicable to the foreign currency notional multiply the below by S0.
        results['analytical_greeks']['spot_delta'] = cp * np.exp(-q * tau) * norm.cdf(cp * d1)
        results['analytical_greeks']['forward_delta'] = cp * norm.cdf(cp * d1)

        # In practice, vega is normalised to measure the change in price for a 1% change in the volatility input
        # Hence we have scaled the analytical vega to a 1% change
        analytical_formula = S0 * np.sqrt(tau) * norm.pdf(d1) * np.exp(-q * tau) # identical for calls and puts
        results['analytical_greeks']['vega'] = analytical_formula * 0.01 # normalised to 1% change

        # Per market convention, theta is normalised to the price decay for 1 calendar day.
        analytical_formula = -(S0*np.exp(-r*tau)*norm.pdf(d1 * cp)*σ)/(2*np.sqrt(tau)) \
            + q * np.exp(-q * tau) * S0 * norm.cdf(d1 * cp) \
            - r * np.exp(-r * tau) * K * norm.cdf(d2 * cp)
        results['analytical_greeks']['theta'] = analytical_formula * 1/365.25 # normalised to 1 calendar day

        # In practice, gamma is  normalised to measure the change in Δ, for a 1% change in the underlying assets price.
        # Hence, we have multiplied the analytical gamma formula by 'S * 0.01'
        analytical_formula = np.exp(-q * tau) * norm.pdf(d1)  / (S0 * σ * np.sqrt(tau)) # identical for calls and puts
        results['analytical_greeks']['gamma'] = (0.01 * S0) * analytical_formula # normalised for 1% change

        # In practice, Rho is normalised to measure the change in price for a 1% change in the underlying interest rate.
        # Hence, we have multiplied the analytical formula result by 0.01 (i.e. 1%)
        analytical_formula = K * tau * np.exp(-q * tau) * norm.cdf(cp * d2)
        results['analytical_greeks']['rho'] = analytical_formula * 0.01


    if numerical_greeks:
        numerical_greeks = pd.DataFrame()

        Δ_shift = 1 / 100  # 1% shift
        σ_shift = 1 / 100  # 1% shift
        θ_shift = 1 / 365.25  # 1 calendar day
        ρ_shift = 1 / 100  # 1% shift

        if F is not None:
            F_upshift, F_downshift = F * (1 + Δ_shift), F * (1 - Δ_shift)
        else:
            F_upshift, F_downshift = None, None

        results_S0_plus = garman_kohlhagen_price(S0=S0*(1+Δ_shift), tau=tau, r_d=r_d, r_f=r_f , cp=cp, K=K, vol=σ, F=F_upshift,analytical_greeks=True)
        X_S_plus, analytical_greeks_S0_plus = results_S0_plus['price'], results_S0_plus['analytical_greeks']
        results_S0_minus = garman_kohlhagen_price(S0=S0*(1-Δ_shift), tau=tau, r_d=r_d, r_f=r_f, cp=cp, K=K, vol=σ, F=F_downshift,analytical_greeks=True)
        X_S_minus, analytical_greeks_S0_minus = results_S0_minus['price'], results_S0_minus['analytical_greeks']

        # Delta, Δ, is the change in an option's price for a small change in the underlying assets price.
        # Δ := ∂X/∂S ≈ (X(S_plus) − X(S_minus)) / (S_plus - S_minus)
        # where X is the option price (whose units is DOM),
        # and S0 is the fx spot price (whose units is DOM/FOR)
        # To get the Δ % applicable to the foreign currency notional multiply the below by S0.
        results['numerical_greeks']['spot_delta'] = (X_S_plus - X_S_minus) / (2 * S0 * Δ_shift)
        results['numerical_greeks']['forward_delta'] = numerical_greeks['spot_delta'] / np.exp(-r_f * tau)

        # Vega, ν, is the change in an options price for a small change in the volatility input
        # ν = ∂X/∂σ ≈ (X(σ_plus) − X(σ_minus)) / (σ_plus - σ_minus).
        # We divide by 0.01 (1%) to normalise the vega to a 1% change in the volatility input, per market convention.
        X_σ_plus = garman_kohlhagen_price(S0=S0, tau=tau, r_d=r_d, r_f=r_f, cp=cp, K=K, vol=σ+σ_shift, F=F)['price']
        X_σ_minus = garman_kohlhagen_price(S0=S0, tau=tau, r_d=r_d, r_f=r_f, cp=cp, K=K, vol=σ-σ_shift, F=F)['price']
        results['numerical_greeks']['vega'] = (X_σ_plus - X_σ_minus) / (2 * (σ_shift / 0.01))

        # Theta, θ, is the change in an options price for a small change in the time to expiry
        # In practice, theta is normalised to measure the change in price, for a 1 day shorter expiry
        # Theta includes
        # 1. Time decay: the change in the option's value as time moves forward, that is, as the expiry date moves closer,
        #               i.e., the time value price tomorrow minus the time value price today.
        # 2. Cost of carry: the interest rate sensitivity that causes the value of the portfolio to change as time progresses,
        #                   e.g., the cost of carry on spots and forwards is included in the theta value.
        X_plus = garman_kohlhagen_price(S0=S0, tau=tau+θ_shift, r_d=r_d, r_f=r_f, cp=cp, K=K, vol=σ, F=F)['price']
        X_minus = garman_kohlhagen_price(S0=S0, tau=tau-θ_shift, r_d=r_d, r_f=r_f, cp=cp, K=K, vol=σ, F=F)['price']
        results['numerical_greeks']['theta'] = (X_minus - X_plus) / 2

        # Gamma, Γ, is the change in an option's delta for a small change in the underlying assets price.
        # Gamma := ∂Δ/∂S ≈ (Δ(S_plus) − Δ(S_minus)) / (S_plus - S_minus).
        results['numerical_greeks']['gamma'] = (analytical_greeks_S0_plus['spot_delta'] - analytical_greeks_S0_minus['spot_delta']) / (2 * (Δ_shift / 0.01))

        # Rho, ρ, is the rate at which the price of an option changes relative to a change in the interest rate.
        # This formulae will yield meaningfully different results on whether the forward rate is an input (or if it is calculated from the interest rate differential)
        X_ρ_plus = garman_kohlhagen_price(S0=S0, tau=tau, r_d=r_d, r_f=r_f+ρ_shift, cp=cp, K=K, vol=σ+σ_shift, F=F)['price']
        X_ρ_minus = garman_kohlhagen_price(S0=S0, tau=tau, r_d=r_d, r_f=r_f+ρ_shift, cp=cp, K=K, vol=σ-σ_shift, F=F)['price']
        results['numerical_greeks']['rho'] = (X_ρ_plus - X_ρ_minus) / 2

    return results


def garman_kohlhagen_solve_strike_from_delta(
        S0: float,
        tau: float,
        r_d: float,
        r_f: float,
        vol: [float, np.ndarray],
        signed_delta: [float, np.ndarray],
        delta_convention: str,
        F: float = None,
        atm_delta_convention: str='forward') -> float:
    """
    Solves the strike for a given Delta-volatility (Δ-σ) quote.

    Parameters
    ----------
    S0 : float
        FX spot rate (specified in # of units of domestic currency per 1 unit of foreign currency).
    tau : float
        Time to expiry (in years).
    r_d : float
        Domestic risk-free interest rate (annualized continuously compounded).
    r_f : float
        Foreign risk-free interest rate (annualized continuously compounded).
    vol : float
        Volatility (annualized standard deviation of FX returns).
    signed_delta : float
        Signed delta, restricted to [0, 0.5] for calls and [-0.5, 0] for puts.
    delta_convention : str
        Delta convention used for the tenors quoting, one of {'regular_spot', 'regular_forward', 'premium_adjusted_spot', 'premium_adjusted_forward'}.
    F : float, optional
        Market forward rate. If None, it will be calculated using interest rate parity (default is None).
    atm_delta_convention : str, optional
        At-the-money delta convention used for the tenors quoting, one of {'forward','per_delta_convention'} (default is 'forward').
        Market convention, is at-the-money delta is quoted in terms of the forward delta, regardless of the delta convention used for the tenors quoting.
    Returns
    -------
    float
        Calculated strike price based on the provided delta.
        
    References
    -------
    [1] Reiswich, Dimitri & Wystup, Uwe. (2010). A Guide to FX Options Quoting Conventions. The Journal of Derivatives. 18. 58-68. 10.3905jod.2010.18.2.058
    """

    # Convert to arrays. Function is vectorised.
    S0, tau, r_f, r_d, σ, Δ = map(lambda x: np.atleast_1d(x).astype(float), (S0, tau, r_f, r_d, vol, signed_delta))

    # Value checks
    assert (S0 > 0.0).all(), S0
    assert (tau > 0.0).all(), tau
    assert (σ > 0.0).all(), σ
    assert (Δ >= -0.5).all() and (Δ <= 0.5).all()
    assert Δ.shape == σ.shape
    cp = np.sign(Δ)
    assert delta_convention in {'regular_spot',
                                'regular_forward',
                                'premium_adjusted_spot',
                                'premium_adjusted_forward'}, delta_convention
    assert atm_delta_convention in {'forward',
                                    'per_delta_convention'}, atm_delta_convention

    if F is None:
        # If market forward rate not supplied, calculate it per interest rate parity
        F = np.atleast_1d(S0 * np.exp((r_d - r_f) * tau))
    else:
        F = np.atleast_1d(F).astype(float)
        assert (F > 0.0).all()

    result = np.zeros(shape=Δ.shape)
    mask_atm = np.abs(Δ) == 0.5
    mask_not_atm = np.logical_not(mask_atm)

    if delta_convention in {'regular_spot','regular_forward'}:

        if delta_convention == 'regular_spot':
            norm_func = norm.ppf(cp * Δ * np.exp(r_f * tau))
        else:
             norm_func = norm.ppf(cp * Δ) # Note: norm.ppf(0.5) = 0. Applicable to for atm-delta-neutral quotes.
        result = (F * np.exp(-cp * norm_func * σ * np.sqrt(tau) + 0.5 * σ**2 * tau))

        if atm_delta_convention == 'per_delta_convention':
            pass
        elif atm_delta_convention == 'forward':
            result[mask_atm] = (F * np.exp(0.5 * σ[mask_atm]**2 * tau))

    elif delta_convention in {'premium_adjusted_spot','premium_adjusted_forward'}:
        
        if np.any(mask_atm):
            # at-the-money Δ-neutral strike, for premium adjusted spot/forward Δ
            result[mask_atm] = (F * np.exp(-1 * 0.5 * σ**2 * tau))[mask_atm]
                       
        if np.any(mask_not_atm):
            for i in range(len(Δ)):
                if Δ[i] != 0.5:
                    # For premium adjusted quotes the strike must be solved numerically. Refer to [1] for full details.

                    def solve_delta(K_):
                        if delta_convention == 'premium_adjusted_spot':
                            multiplier = np.exp(-r_f * tau)
                        else:
                            multiplier = 1
                        return (multiplier * (cp[i] * K_ / F) * norm.cdf(cp[i] * (np.log(F/K_) - 0.5 * σ[i]**2 * tau) / (σ[i] * np.sqrt(tau)))) - Δ[i]

                    # Solve the upper bound, 'K_max' for the numerical solver
                    # The strike of a premium adjusted Δ-σ quote is ALWAYS below the regular (non premium-adjusted) Δ-σ quote
                    # Hence, we analytically calculate the K, assuming the σ was a regular Δ quote
                    delta_convention_adj = delta_convention.replace('premium_adjusted','regular')
                    K_max = garman_kohlhagen_solve_strike_from_delta(
                        S0=S0,tau=tau, r_d=r_d, r_f=r_f, vol=σ[i], signed_delta=Δ[i], delta_convention=delta_convention_adj, F=F).item()
    
                    # Put Option
                    if Δ[i] < 0: 
                        solution = root_scalar(solve_delta, x0=F, bracket=[0.000001, K_max])
                        
                    # Call Option
                    if Δ[i] > 0: 
                        # For the premium adjusted call Δ, due to non-monotonicity, two strikes can be solved numerically for call options (but not put options).
                        # To avoid this, we solve a lower bound, 'K_min' to guarantee we get the correct solution in the numerical solver.
                        # The lower bound is the 'maximum' Δ, hence we numerically solve the maximum Δ
                        def solve_k_lower_bound(K): # TODO test and remove cp, F
                            d1 = (np.log(S0 / K) + (r_d - r_f + 0.5 * σ[i]**2) * tau) / (σ[i] * np.sqrt(tau))
                            d2 = d1 - σ[i] * np.sqrt(tau)
                            return σ[i] * np.sqrt(tau) * norm.cdf(d2) - norm.pdf(d2)
                        
                        try: 
                            solution = root_scalar(solve_k_lower_bound, x0=F ,bracket=[0.00001, K_max])
                        except ValueError('the numerical solve for K_min, for delta', Δ[i], ', resulted in an error'):
                            pass
                    
                        if solution.converged:
                            K_min = solution.root
                        else:
                            raise ValueError('the numerical solver for K_min, for delta', Δ[i], ', did not converge')
                        
                        try:
                            solution = root_scalar(solve_delta, x0=K_max, bracket=[K_min,K_max])
                        except ValueError('the numerical solve for premium adjusted strike for delta', Δ[i], ', resulted in an error'):
                            print('This is likely due to an error in the input, for example typos or specifying the delta_convention as spot-delta when it is actually forward-delta')
                        
                    if solution.converged:
                        result[i] = solution.root
                    else:
                        raise ValueError('the numerical solver for the premium adjusted strike for delta', Δ[i], ', did not converge')
    else:
        raise ValueError("'delta_convention' must be one of {regular_spot, regular_forward, premium_adjusted_spot, premium_adjusted_forward}", delta_convention)


    return result
                    

def garman_kohlhagen_solve_implied_vol(
        S0: float,
        tau: float,
        r_d: float,
        r_f: float,
        cp: int,
        K: float,
        X: float,
        vol_guess: float) -> float:
    """
    Solve for the implied volatility using the Garman-Kohlhagen model for a European Vanilla FX option.

    This function attempts to compute the implied volatility by matching the observed option price 
    with the Garman-Kohlhagen model price. It first uses Newton's method for faster convergence 
    and falls back to Brent's method if necessary.

    Parameters
    ----------
    S0 : float
        FX spot price (specified in # of units of domestic currency per 1 unit of foreign currency).
    tau : float
        Time to expiry in years.
    r_d : float
        Domestic risk-free interest rate (annualized continuously compounded).
    r_f : float
        Foreign risk-free interest rate (annualized continuously compounded).
    cp : int
        Option type: 1 for call, -1 for put.
    K : float
        Strike price of the option (specified in # of units of domestic currency per 1 unit of foreign currency).
    X : float
        Observed market option price.
    vol_guess : float
        Initial guess for the volatility.

    Returns
    -------
    float
        The implied volatility, or np.inf if no solution is found.

    Raises
    ------
    RuntimeError
        If Newton's method fails to converge within the iteration limit.
    ValueError
        If Brent's method fails to find a solution within the specified bracket.

    Notes
    -----
    - The function first tries Newton's method, which is faster but less robust, with a tolerance of 1e-4.
    - If Newton's method fails, Brent's method is used for fallback, searching between 0.0001 and 2.
    - If neither method succeeds, the function returns np.inf to indicate failure.
    """
    def error_function(vol_):
        return garman_kohlhagen_price(S0=S0,tau=tau,r_d=r_d,r_f=r_f,cp=cp,K=K,vol=vol_)['price']  - X

    try:
        # Try Netwon's method first (it's faster but less robust)
        return newton(lambda vol_: error_function(vol_), x0=vol_guess, tol=1e-4, maxiter=50).item()
    except RuntimeError:
        # Fallback to Brent's method
        try:
            return root_scalar(lambda vol_: error_function(vol_), bracket=[0.0001, 2], method='brentq').root
        except ValueError:
            return np.inf















# -*- coding: utf-8 -*-
import os
import numpy as np
import pandas as pd
from scipy.stats import norm
from scipy.optimize import root_scalar, minimize

if __name__ == "__main__":
    os.chdir(os.environ.get('PROJECT_DIR_FRM'))

VOL_SLN_BOUNDS = (0.1 / 100, 1000 / 100)  # 0.1% to 1000% (0.001 to 10)
VOL_N_BOUNDS = (0.01 / 100, 100 / 100)  # 0.01% to 100% (0.0001 to 1)

def black76_price(
        F: [float, np.array],
        tau: [float, np.array],
        cp: [float, np.array],
        K: [float, np.array],
        vol_sln: [float, np.array],
        ln_shift: [float, np.array],
        annuity_factor: [float, np.array]=1,
        intrinsic_time_split: bool=False,
        analytical_greeks: bool=False,
        numerical_greeks: bool=False):
    """
    Black76 pricing + greeks.

    The function has the parameters 'annuity_factor' instead of a risk-free rate.
    This adjustment allows for more generic applications of Black76 to caplets/floorlets and swaptions instead of just European options delivered at expiry.

    Parameters
    ----------
    F : float
        Forward price.
    tau : float
        Time to expiry (in years).
    K : float
        Strike price.
    cp : int
        Option type: 1 for call option, -1 for put option.
    vol_sln : float
        Volatility (annualized).
    ln_shift : float, optional
        Log-normal shift, applied to forward price and strike (default is 0).
    annuity_factor : float, optional
        Multiplier to adjust the Black76 forward price to present value (default is 1).
        This is composed of the discount factor and the accrual period fraction.
    analytical_greeks : bool, optional
        If True, analytical greeks will be calculated and returned (default is False).
    numerical_greeks : bool, optional
        If True, numerical greeks will be calculated and returned using finite differences (default is False).
    intrinsic_time_split : bool, optional
        If True, splits option value into intrinsic and time value components (default is False).

    Returns
    -------
    results : dict
          Dictionary containing the following key-value pairs:
        - 'price' : float
            Option price.
        - 'analytical_greeks' : pd.DataFrame
            Analytical greeks; columns: 'delta', 'vega', 'theta', 'gamma', 'rho'.
        - 'numerical_greeks' : pd.DataFrame
            Numerical greeks; columns: 'delta', 'vega', 'theta', 'gamma', 'rho'.
        - 'intrinsic_time_split' : float
            Dictionary of intrinsic and time value components; 'intrinsic', 'time'.
    """

    # Convert to arrays. Function is vectorised.
    F, tau, cp, K, σB, ln_shift, annuity_factor = map(np.atleast_1d, (F, tau, cp, K, vol_sln, ln_shift, annuity_factor))

    # Validation checks
    assert F.shape == tau.shape
    shapes = set([param.shape for param in (F, tau, cp, K, σB, ln_shift, annuity_factor)])
    assert len(shapes) in [1, 2]
    if len(shapes) == 2:
        assert (1,) in shapes

    F = F + ln_shift
    K = K + ln_shift

    # Price per Black76 formula
    d1 = (np.log(F/K) + (0.5 * σB**2 * tau)) / (σB*np.sqrt(tau))
    d2 = d1 - σB*np.sqrt(tau)
    X = annuity_factor * cp * (F * norm.cdf(cp * d1) - K * norm.cdf(cp * d2))

    results = {'price': X}

    # Intrinsic / time value split
    if intrinsic_time_split:
        intrinsic = np.full_like(X, np.nan)
        intrinsic[tau>0] = np.maximum(0, cp * (F - K))[tau>0]
        intrinsic[tau==0] = X[tau==0]
        results['intrinsic'] = intrinsic
        results['time'] = X - intrinsic

    if analytical_greeks:
        results['analytical_greeks'] = pd.DataFrame()

        results['analytical_greeks']['delta'] = cp * norm.cdf(cp * d1) * annuity_factor

        analytical_vega = F * np.sqrt(tau) * norm.pdf(d1)
        # In practice, vega is displayed as normalised to a 1% shift, hence the 0.01 multiplier.
        results['analytical_greeks']['vega'] = 0.01 * analytical_vega *  annuity_factor

    if numerical_greeks:
        results['numerical_greeks'] = pd.DataFrame()

        # Delta, Δ, is the change in an option's price for a small change in the underlying assets (forward) price.
        # Δ := ∂X/∂F ≈ (X(F_plus) − X(F_minus)) / (F_plus - F_minus)
        Δ_shift = 0.01
        results_F_plus = black76_price(F=F*(1+Δ_shift), tau=tau, K=K, cp=cp, vol_sln=vol_sln, ln_shift=ln_shift,
                           annuity_factor=annuity_factor, analytical_greeks=True)
        results_F_minus = black76_price(F=F*(1-Δ_shift), tau=tau, K=K, cp=cp, vol_sln=vol_sln, ln_shift=ln_shift,
                            annuity_factor=annuity_factor, analytical_greeks=True)
        results['numerical_greeks']['delta'] = (results_F_plus['price'] - results_F_minus['price']) / (2 * F * Δ_shift)

        # Vega, ν, is the change in an options price for a small change in the volatility input
        # ν = ∂X/∂σ ≈ (X(σ_plus) − X(σ_minus)) / (σ_plus - σ_minus).
        # We divide by 0.01 (1%) to normalise the vega to a 1% change in the volatility input, per market convention.
        σ_shift = 0.01
        results_σ_plus = black76_price(F=F, tau=tau, K=K, cp=cp, vol_sln=vol_sln+σ_shift, ln_shift=ln_shift,
                           annuity_factor=annuity_factor, analytical_greeks=True)
        results_σ_minus = black76_price(F=F, tau=tau, K=K, cp=cp, vol_sln=vol_sln-σ_shift, ln_shift=ln_shift,
                            annuity_factor=annuity_factor, analytical_greeks=True)
        results['numerical_greeks']['vega'] = (results_σ_plus['price'] - results_σ_minus['price']) / (2 * (σ_shift / 0.01))

    return results


def black76_solve_implied_vol(
        F: [float, np.array],
        tau: [float, np.array],
        cp: [float, np.array],
        K: [float, np.array],
        ln_shift: [float, np.array],
        X: float,
        vol_sln_guess: float=0.1,
        annuity_factor: [float, np.array]=1,
        ) -> float:
    """Solve the implied normal volatility with the Black76 pricing formula."""

    def error_function(vol_):
        # Return the errors (relative to the price, as we want invariance to the price) to be minimised.
        # For gradient-based optimisation methods, return squared error to ensure differentiability.
        relative_error = (black76_price(F=F, tau=tau, cp=cp, K=K, vol_sln=vol_, ln_shift=ln_shift, annuity_factor=annuity_factor)['price'].sum() - X) / X
        error = relative_error**power
        if abs(error) < obj_func_tol**power:  # Check against objective function tolerance
            raise StopIteration(vol_)  # Use StopIteration to terminate optimization early if tolerance is met
        return error

    # Want prices to be within 0.001% - i.e. if price is 100,000, acceptable range is (99999, 100001)
    # Hence set obj function tol to (0.001%)**2 = 1e-10 (due to the squaring of the relative error in the obj function)
    xtol = 1e-6
    obj_func_tol = 1e-5
    x0 = vol_sln_guess
    bounds = VOL_SLN_BOUNDS

    # Note Brent's does not work for a SSE obj. function as f(a) and f(b) must have different signs.
    # f(a), f(b) are the function values of the bracket.
    try:
        power = 1 # Use linear error for the root_scalar method
        res = root_scalar(lambda vol_: error_function(vol_), x0=x0, bracket=bounds, xtol=xtol, method='brentq')
        if abs(error_function(res.root)) < obj_func_tol**power:
            return res.root
        else:
            raise RuntimeError
    except StopIteration as e:
        # Return the root if we meet the objective tolerance early
        return e.value
    except RuntimeError:
        try:
            # Fallback to L-BFGS-B if root_scalar fails
            options = {'ftol': obj_func_tol, 'gtol': 0} # gtol set to 0, so optimisation is terminated based on ftol
            power = 2 # Use squared error for the optimisation
            res = minimize(fun=error_function, x0=np.atleast_1d(x0), bounds=[bounds], method='L-BFGS-B', options=options)
            # 2nd condition required as we have overridden options to terminate based on ftol
            if res.success or abs(res.fun) < obj_func_tol**power:
                return res.x[0]
            else:
                raise ValueError('Optimisation to solve normal volatility did not converge.')
        except StopIteration as e:
            # Return the root if we meet the objective tolerance early
            return e.value[0]


def bachelier_price(
        F: [float, np.array],
        tau: [float, np.array],
        cp: [float, np.array],
        K: [float, np.array],
        vol_n: [float, np.array],
        annuity_factor: [float, np.array] = 1,
        intrinsic_time_split: bool=False,
        analytical_greeks: bool=False,
        numerical_greeks: bool=False):
    """
    Bachelier pricing + greeks.

    The function has the parameters 'annuity_factor' instead of a risk-free rate.
    This adjustment allows for more generic applications of Bachelier to caplets/floorlets and swaptions instead of just European options delivered at expiry.

    Parameters
    ----------
    F : float
        Forward price.
    tau : float
        Time to expiry (in years).
    cp : int
        Option type: 1 for call option, -1 for put option.
    K : float
        Strike price.
    vol_n : float
        Volatility (annualized).
    annuity_factor : float, optional
        Multiplier to adjust the Black76 forward price to present value (default is 1).
        This is composed of the discount factor and the accrual period fraction.
    intrinsic_time_split : bool, optional
        If True, splits option value into intrinsic and time value components (default is False).
    analytical_greeks : bool, optional
        If True, analytical greeks will be calculated and returned (default is False).
    numerical_greeks : bool, optional
        If True, numerical greeks will be calculated and returned using finite differences (default is False).

    Returns
    -------
    results : dict
          Dictionary containing the following key-value pairs:
        - 'price' : float
            Option price.
        - 'analytical_greeks' : float
            Dictionary of analytical greeks; 'delta', 'vega', 'theta', 'gamma', 'rho'.
        - 'numerical_greeks' : float
            Dictionary of numerical greeks; 'delta', 'vega', 'theta', 'gamma', 'rho'.
        - 'intrinsic_time_split' : float
            Dictionary of intrinsic and time value components; 'intrinsic', 'time'.
    """

    # Convert to arrays. Function is vectorised.
    F, tau, cp, K, σN, annuity_factor = map(np.atleast_1d, (F, tau, cp, K, vol_n, annuity_factor))

    # Validation checks
    assert F.shape == tau.shape
    shapes = set([param.shape for param in (F, tau, cp, K, σN, annuity_factor)])
    assert len(shapes) in [1, 2]
    if len(shapes) == 2:
        assert (1,) in shapes

    # Price per Bachelier formula
    d = (F - K) / (σN * np.sqrt(tau))
    X = annuity_factor * ( cp * (F - K) * norm.cdf(cp * d) + σN * np.sqrt(tau) * norm.pdf(d) )

    results = {'price': X}

    # Intrinsic / time value split
    if intrinsic_time_split:
        intrinsic = np.full_like(X, np.nan)
        intrinsic[tau>0] = np.maximum(0, cp * (F - K))[tau>0]
        intrinsic[tau==0] = X[tau==0]
        results['intrinsic'] = intrinsic
        results['time'] = X - intrinsic

    if analytical_greeks:
        results['analytical_greeks'] = pd.DataFrame()

        results['analytical_greeks']['delta'] = cp * norm.cdf(cp * d) * annuity_factor

        # Market convention is to adjust normal vega to a 1 basis point impact hence the 0.0001 multiplier.
        results['analytical_greeks']['vega'] = np.sqrt(tau) * norm.pdf(cp * d) * annuity_factor * 0.0001

        # Normalised to 1 calendar day.
        results['analytical_greeks']['theta'] = (-0.5 * norm.pdf(d) * σN / np.sqrt(tau)) * annuity_factor / (1/365.25)

    return results


def bachelier_solve_implied_vol(
        F: [float, np.array],
        tau: [float, np.array],
        cp: [float, np.array],
        K: [float, np.array],
        X: float,
        vol_n_guess: float=0.01,
        annuity_factor: [float, np.array]=1,
        ) -> float:
    """Solve the implied normal volatility with the Bachelier pricing formula."""

    def error_function(vol_):
        # Return the errors (relative to the price, as we want invariance to the price) to be minimised.
        # For gradient-based optimisation methods, return squared error to ensure differentiability.
        relative_error = (bachelier_price(F=F, tau=tau, cp=cp, K=K, vol_n=vol_, annuity_factor=annuity_factor)['price'].sum() - X) / X
        error = relative_error**power
        if abs(error) < obj_func_tol**power:  # Check against objective function tolerance
            raise StopIteration(vol_)  # Use StopIteration to terminate optimization early if tolerance is met
        return error

    # Want prices to be within 0.001% - i.e. if price is 100,000, acceptable range is (99999, 100001)
    # Hence set obj function tol to (0.001%)**2 = 1e-10 (due to the squaring of the relative error in the obj function)
    xtol = 1e-6
    obj_func_tol = 1e-5
    x0 = vol_n_guess
    bounds = VOL_N_BOUNDS

    # Note Brent's does not work for a SSE obj. function as f(a) and f(b) must have different signs.
    # f(a), f(b) are the function values of the bracket.
    try:
        power = 1 # Use linear error for the root_scalar method
        res = root_scalar(lambda vol_: error_function(vol_), x0=x0, bracket=bounds, xtol=xtol, method='brentq')
        if abs(error_function(res.root)) < obj_func_tol**power:
            return res.root
        else:
            raise RuntimeError
    except StopIteration as e:
        # Return the root if we meet the objective tolerance early
        return e.value
    except RuntimeError:
        try:
            # Fallback to L-BFGS-B if root_scalar fails
            options = {'ftol': obj_func_tol, 'gtol': 0} # gtol set to 0, so optimisation is terminated based on ftol
            power = 2 # Use squared error for the optimisation
            res = minimize(fun=error_function, x0=np.atleast_1d(x0), bounds=[bounds], method='L-BFGS-B', options=options)
            # 2nd condition required as we have overridden options to terminate based on ftol
            if res.success or abs(res.fun) < obj_func_tol**power:
                return res.x[0]
            else:
                raise ValueError('Optimisation to solve normal volatility did not converge.')
        except StopIteration as e:
            # Return the root if we meet the objective tolerance early
            return e.value[0]



def black76_sln_to_normal_vol_analytical(
        F: float,
        tau: float,
        K: float,
        vol_sln: float,
        ln_shift: float
    ):
    """
    Calculates the normal volatility from the Black76 log-normal volatility.

    Parameters
    ----------
    F : float
        Forward price.
    tau : float
        Time to expiry (in years).
    K : float
        Strike price.
    vol_sln : float
        Log-normal volatility (annualized).
    ln_shift : float, optional
        Log-normal shift, applied to forward price and strike (default is 0).

    Returns
    -------
    float
        Normal volatility (annualized).

    References:
    [1] Hagan, Patrick & Lesniewski, Andrew & Woodward, Diana. (2002). Managing Smile Risk. Wilmott Magazine. 1. 84-108.
    """
    F = F + ln_shift
    K = K + ln_shift
    σB = vol_sln

    # Per B.63. in reference [1].
    # σN = σB * (F - K) * (1 - σB**2 * tau / 24) / np.log(F / K)

    # Per B.64 in reference [1]. Slightly more accurate than B.63.
    ln_F_K = np.log(F / K)
    σN = σB * np.sqrt(F*K) \
         * (1 + (1/24) * ln_F_K**2 + 1/1920 * ln_F_K**4) \
         / (1 + (1/24) * (1 - (1/120) * ln_F_K**2) * σB**2 * tau + (1/5760) * σB**4 * tau**2)
    return σN


def black76_sln_to_normal_vol(
        F: float,
        tau: float,
        K: float,
        vol_sln: float,
        ln_shift: float
    ) -> float:

    # If the forward and strike are equal, we can equate the Black76 & Bachelier formulae, and solve analytically.
    if abs(F - K) < 1e-10:
        F = F + ln_shift # K is not used as K=F
        res = (2.0 * F * norm.cdf((vol_sln * np.sqrt(tau)) / 2.0) - F) / (np.sqrt(tau) * norm.pdf(0))
    else:
        # If needed, faster method detailed in: Le Floc'h, Fabien, Fast and Accurate Analytic Basis Point Volatility (April 10, 2016).
        # This solve is invariant to the call/put perspective and the risk-free rate / annuity factor.
        cp = 1
        black76_px = black76_price(F=F, tau=tau, K=K, cp=cp, vol_sln=vol_sln, ln_shift=ln_shift)['price'][0]
        vol_n_guess = np.atleast_1d(black76_sln_to_normal_vol_analytical(F=F, tau=tau, K=K, vol_sln=vol_sln, ln_shift=ln_shift))
        res = bachelier_solve_implied_vol(F=F, tau=tau, cp=cp, K=K, X=black76_px, vol_n_guess=vol_n_guess.item())
    return np.atleast_1d(res).item()


def shift_black76_vol(
        F: [float, np.float64],
        tau: [float, np.float64],
        K: [float, np.float64],
        vol_sln: [float, np.float64],
        from_ln_shift: [float, np.float64],
        to_ln_shift: [float, np.float64]
    ):

    # The solve is invariant to the call/put perspective, risk-free rate/annuity factor.
    cp = 1

    black76_px = black76_price(F=F, tau=tau, K=K, cp=cp, vol_sln=vol_sln, ln_shift=from_ln_shift)['price'][0]

    def obj_func_relative_px_error(vol_new_ln_shift):
        # Return the square of the relative error (to the price, as we want invariance to the price) to be minimised.
        # Squared error ensures differentiability which is critical for gradient-based optimisation methods.
        return ((black76_price(F=F, tau=tau, K=K, cp=cp, vol_sln=vol_new_ln_shift, ln_shift=to_ln_shift)['price'][0] - black76_px) / black76_px)**2

    # Want prices to be within 0.001% - i.e. if price is 100,000, acceptable range is (99999, 100001)
    # Hence set obj function tol 'ftol' to (0.001%)**2 = 1e-10 (due to the squaring of the relative error in the obj function)
    # We set gtol to zero, so that the optimisation is terminated based on ftol.
    obj_func_tol = 1e-5 ** 2  # 0.001%^2
    options = {'ftol': obj_func_tol, 'gtol': 0}
    res = minimize(fun=obj_func_relative_px_error,
                   x0=np.atleast_1d(vol_sln * (F / (F + (to_ln_shift - from_ln_shift)))),
                   bounds=[VOL_SLN_BOUNDS],
                   method='L-BFGS-B',
                   options=options)
    # 2nd condition required as we have overridden options to terminate based on ftol
    if res.success or abs(res.fun) < obj_func_tol:
        return res.x[0]
    else:
        # Further development could include:
        # (i) fallback to other optimisation methods (e.g. 'trust-constr')
        # (ii) addition of 'xtol' parameter into the optimisation (it is a parameter that is not available in 'L-BFGS-B')
        #      'xtol' allows termination of the optimisation when the solved parameter (i.e. vol_sln) only changes by a small amount (e.g. 0.001%)
        vol_new_ln_shift = res.x[0]
        black76_px_shifted = black76_price(F=F, tau=tau, K=K, cp=cp, vol_sln=vol_new_ln_shift, ln_shift=to_ln_shift)['price'][0]
        print({'F': F, 'tau': tau, 'K': K, 'vol_n': vol_sln, 'ln_shift': from_ln_shift})
        print({'black76_px': black76_px, 'black76_px_shifted': black76_px_shifted, 'vol_new_ln_shift': vol_new_ln_shift,
               'relative_error': obj_func_relative_px_error(vol_sln)})
        print(res)
        raise ValueError('Optimisation to shift Black76 volatility did not converge.')


def normal_vol_atm_to_black76_sln_atm(
        F: [float, np.float64, np.array],
        tau: [float, np.float64, np.array],
        vol_n_atm: [float, np.float64, np.array],
        ln_shift: [float, np.float64, np.array]):
    F = F + ln_shift
    return (2.0 / np.sqrt(tau)) * norm.ppf((vol_n_atm * np.sqrt(tau) * norm.pdf(0) + F) / (2.0 * F))


def normal_vol_to_black76_sln(
        F: [float, np.float64],
        tau: [float, np.float64],
        K: [float, np.float64],
        vol_n: [float, np.float64],
        ln_shift: [float, np.float64]
    ) -> float:
    # If the forward and strike are equal, we use an analytical solution from equating the Black76 & Bachelier formulae
    if abs(F - K) < 1e-10:
        normal_vol_atm_to_black76_sln_atm(F=F, tau=tau, vol_n_atm=vol_n, ln_shift=ln_shift)

    # The solve is invariant to the risk-free rate or the call/put perspective.
    cp = 1
    bachelier_px = bachelier_price(F=F, tau=tau, K=K, cp=cp, vol_n=vol_n)['price'][0]

    def obj_func_relative_px_error(vol_sln):
        # Return the square of the relative error (to the price, as we want invariance to the price) to be minimised.
        # Squared error ensures differentiability which is critical for gradient-based optimisation methods.
        return ((bachelier_px - black76_price(F=F, tau=tau, K=K, cp=cp, vol_sln=vol_sln, ln_shift=ln_shift)['price'][0]) / bachelier_px)**2

    # Want prices to be within 0.001% - i.e. if price is 100,000, acceptable range is (99999, 100001)
    # Hence set obj function tol 'ftol' to (0.001%)**2 = 1e-10 (due to the squaring of the relative error in the obj function)
    # We set gtol to zero, so that the optimisation is terminated based on ftol.
    obj_func_tol = 1e-5 ** 2  # 0.001%^2
    options = {'ftol': obj_func_tol, 'gtol': 0}
    res = minimize(fun=obj_func_relative_px_error,
                   x0=np.atleast_1d(vol_n / (F + ln_shift)),
                   bounds=[VOL_SLN_BOUNDS],
                   method='L-BFGS-B',
                   options=options)
    # 2nd condition required as we have overridden options to terminate based on ftol
    if res.success or abs(res.fun) < obj_func_tol:
        return res.x[0]
    else:
        # Further development could include:
        # (i) fallback to other optimisation methods (e.g. 'trust-constr')
        # (ii) addition of 'xtol' parameter into the optimisation (it is a parameter that is not available in 'L-BFGS-B')
        #      'xtol' allows termination of the optimisation when solved parameter (i.e. vol_sln) only changes by a small amount (e.g. 0.001%)
        vol_sln = res.x[0]
        black76_px = black76_price(F=F, tau=tau, K=K, cp=cp, vol_sln=vol_sln, ln_shift=ln_shift)['price'][0]
        print({'F': F, 'tau': tau, 'K': K, 'vol_n': vol_n, 'ln_shift': ln_shift})
        print({'bachelier_px': bachelier_px, 'black76_px': black76_px, 'vol_sln': vol_sln,
               'relative_error': obj_func_relative_px_error(vol_sln)})
        print(res)
        raise ValueError('Optimisation to convert normal volatility to log-normal volatility did not converge.')






# -*- coding: utf-8 -*-
import os
if __name__ == "__main__":
    os.chdir(os.environ.get('PROJECT_DIR_FRM')) 

import multiprocessing
from concurrent.futures import ProcessPoolExecutor
import numpy as np
import time

NUMBER_OF_CORES = 8
APPLY_CONTROL_VARIATES = True
APPLY_PARALLEL_PROCESSING = False

x0 = 10
dt = 1/365
drift = 0.05
speed = 0.3

nb_timesteps = 10
nb_assets = 1
nb_iterations = 1000 * 1000
chunk_size = nb_iterations // NUMBER_OF_CORES


def gbm_helper(start_iter, end_iter):
    np.random.seed()
    sqrt_dt = np.sqrt(dt)
    local_nb_iterations = end_iter - start_iter
    
    if APPLY_CONTROL_VARIATES:
        # https://en.wikipedia.org/wiki/Control_variates
        W_1 = np.random.randn(int(nb_timesteps), int(nb_assets), int(local_nb_iterations // 2))
        result_1 = x0 * np.cumprod(np.exp((drift - 0.5 * speed**2) * dt + speed * sqrt_dt * W_1), axis=1)
        W_2 = -W_1
        result_2 = x0 * np.cumprod(np.exp((drift - 0.5 * speed**2) * dt + speed * sqrt_dt * W_2), axis=1)
        result = np.append(result_1, result_2, axis=2)
    else:
        W = np.random.randn(nb_timesteps, nb_assets, local_nb_iterations)
        result = x0 * np.cumprod(np.exp((drift - 0.5 * speed**2) * dt + speed * sqrt_dt * W), axis=1)

    return result


if __name__ == '__main__':
    
    
    print('# of cores:', multiprocessing.cpu_count())
    
    
    t1 = time.time()
    
    with ProcessPoolExecutor(max_workers=NUMBER_OF_CORES) as executor:
        future_results = [executor.submit(gbm_helper, i, i + chunk_size) for i in range(0, nb_iterations, chunk_size)]
        results = [future.result() for future in future_results]
        final_result = np.concatenate(results, axis=2)

    print(final_result[1,-1,:].mean())
    t2 = time.time()
    print(t2-t1)    



    final_result = gbm_helper(0, nb_iterations)
        
    print(final_result[1,-1,:].mean())
    t3 = time.time()
    print(t3-t2)


# Talk with Ruchita
# 1. https://en.wikipedia.org/wiki/Antithetic_variates - can we do this for GBM simulations?
# 2. Any other variance reduction techneques applicable?
# 3. Is there a way we can check the simuluation converged? 
# 
#   - is there a magic number i.e 1m iterations 
#   - note this example is a simple example and will be used as the base for more complex option valuations
# My parallel processing isn't much quicker, I was hoping for a improvement similiar to the number of cores?

# How would this be built in effeciently and could it consider parallel processing?



# -*- coding: utf-8 -*-
import datetime as dt
import numpy as np
import os
import pandas as pd
from scipy.interpolate import interp1d
import time
from typing import Optional

from frm.enums import DayCountBasis
from frm.utils import clean_tenor, tenor_to_date_offset, year_frac, get_busdaycal, workday


def cash_rates_quote_helper(
        df: pd.DataFrame,
        curve_date: pd.Timestamp,
        day_count_basis: DayCountBasis,
        settlement_delay: Optional[int]=None,
        settlement_date: Optional[pd.Timestamp]=None,
        cal: np.busdaycalendar=np.busdaycalendar):

    assert 'rate' in df.columns
    assert 'tenor' in df.columns or ('effective_date' in df.columns and 'maturity_date' in df.columns)

    df['tenor'] = df['tenor'].apply(clean_tenor)

    if settlement_date is None:
        assert settlement_delay is not None
        settlement_date = workday(dates=curve_date, offset=settlement_delay, cal=cal)

    if 'effective_date' not in df.columns:
        df['has_settlement_delay'] = df['tenor'] != 'on'
        df.loc[~df['has_settlement_delay'],'effective_date'] = curve_date
        df.loc[df['has_settlement_delay'],'effective_date'] = settlement_date

    if 'maturity_date' not in df.columns:
        date_offset = df['tenor'].apply(tenor_to_date_offset).values
        df['maturity_date'] = workday(dates=(df['effective_date'] + date_offset), offset=0, cal=cal)

    df['period_years'] = year_frac(df['effective_date'], df['maturity_date'], day_count_basis)

    return df

def bootstrap_cash_rates(
        df: pd.DataFrame,
        curve_date: pd.Timestamp,
        day_count_basis_years: DayCountBasis=DayCountBasis.ACT_ACT):

    df['effective_date_years'] = year_frac(curve_date, df['effective_date'], day_count_basis_years)
    df['maturity_date_years'] = year_frac(curve_date, df['maturity_date'], day_count_basis_years)
    df['effective_date_discount_factor'] = np.nan
    df['maturity_date_discount_factor'] = np.nan

    df.sort_values(by='maturity_date', inplace=True)
    for i, row in df.iterrows():
        if not row['has_settlement_delay']:
            df.at[i, 'effective_date_discount_factor'] = 1.0
            df.at[i, 'maturity_date_discount_factor'] = 1.0 / (1 + row['rate'] * row['period_years'])
        else:
            mask = df['maturity_date'] <= row['effective_date']

            x = row['effective_date_years']
            X = df.loc[mask, 'maturity_date_years'].values
            Y = np.log(df.loc[mask, 'maturity_date_discount_factor'].values)
            interp_func = interp1d(X, Y, kind='linear')
            df.at[i, 'effective_date_discount_factor'] = np.exp(interp_func(x))

            df.at[i, 'maturity_date_discount_factor'] = df.at[i, 'effective_date_discount_factor'] \
                                                                / (1 + row['rate'] * row['period_years'])

    #df = df.drop(columns=['effective_date_years', 'maturity_date_years'], errors='ignore')

    return df



def bootstrap_futures(
        df: pd.DataFrame,
        curve_date: pd.Timestamp,
        cash_rates_df: pd.DataFrame,
        day_count_basis: DayCountBasis,
        day_count_basis_years: DayCountBasis=DayCountBasis.ACT_ACT):

    if 'futures_rate' not in df.columns:
        df['futures_rate'] = (100 - df['price']) / 100
    if 'convexity_bps' not in df.columns:
        df.insert(df.columns.get_loc('price') + 1, 'convexity_bps', 0)
    if 'forward_rate' not in df.columns:
        df['forward_rate'] = df['futures_rate'] - df['convexity_bps'] / 10000

    df['period_years'] = year_frac(df['effective_date'], df['maturity_date'], day_count_basis)
    df['effective_date_years'] = year_frac(curve_date, df['effective_date'], day_count_basis_years)
    df['maturity_date_years'] = year_frac(curve_date, df['maturity_date'], day_count_basis_years)

    X = cash_rates_df['maturity_date_years'].values
    Y = np.log(cash_rates_df['maturity_date_discount_factor'].values)

    df.sort_values(by='effective_date', inplace=True)
    df.reset_index(inplace=True, drop=True)
    for i, row in df.iterrows():
        x = row['effective_date_years']
        interp_func = interp1d(X, Y, kind='linear', fill_value='extrapolate')
        df.at[i, 'effective_date_discount_factor'] = np.exp(interp_func(x))

        df.at[i, 'maturity_date_discount_factor'] = df.at[i, 'effective_date_discount_factor'] \
                                                    / (1 + row['forward_rate'] * row['period_years'])

        X = np.append(X, row['maturity_date_years'])
        Y = np.append(Y, np.log(df.at[i, 'maturity_date_discount_factor']))

    return df


def futures_helper_asx_90day(
        df: pd.DataFrame
    ):
    # Helper function to setup contract details for ASX 90-day bank bill futures.
    # https://www.asx.com.au/content/dam/asx/participants/derivatives-market/90-Day-bank-bill-futures-factsheet.pdf

    # From the contract code, index the effective period of the BBSW 3M fixing of the contract.

    if 'months' not in df.columns or 'years' not in df.columns:

        if 'ticker' in df.columns:

            df['ticker'] = df['ticker'].str.upper().str.strip()
            contract_code = df['ticker'].str[:2]
            assert (contract_code == 'IR').all() # 'IR' for 90-day bank bill futures

            futures_month_code = df['ticker'].str[2]
            assert futures_month_code.isin(['H', 'M', 'U', 'Z']).all() # 'H', 'M', 'U', 'Z' for Mar, Jun, Sep, Dec
            df['months'] = futures_month_code.map({'H': 3, 'M': 6, 'U': 9, 'Z': 12})

            assert df['ticker'].str.len().isin([1, 2, 4]).all()
            futures_year_code = df['ticker'].str[3:].astype(int)  # 1-4-digit year code

            def resolve_year(year_code):
                year_code_str = str(year_code)
                if len(year_code_str) == 1:
                    # Map single-digit year codes to 2020-2029
                    # TODO: does not support cases where the futures schedule spans multiple decades. i.e from 2028-2031.
                    return 2020 + int(year_code_str)
                elif len(year_code_str) == 2:
                    # Map two-digit year codes to current century
                    return 2000 + int(year_code_str)
                elif len(year_code_str) == 4:
                    # Use four-digit year codes as is
                    return int(year_code_str)
                else:
                    raise ValueError(f'Unexpected year code length: {year_code_str}')

            df['years'] = futures_year_code.apply(resolve_year)

        elif 'month_year' in df.columns:
            # SEP23, DEC23, MAR24, JUN24...
            # SEP2024, DEC2024, MAR2025, JUN2025...
            pass



    expiry_dates, settlement_dates = [], []
    for years in range(df['years'].min(), df['years'].max()+1):
        for months in [3, 6, 9, 12]:
            if ((df['years'] == years) & (df['months'] == months)).any():
                # Settlement Day = 2nd Friday of each delivery month (Mar, Jun, Sep, Dec).
                # Start from the 1st of the month and get all Fridays in the month
                month_start = pd.Timestamp(year=years, month=months, day=1)
                fridays = [d for d in pd.date_range(month_start, month_start + pd.offsets.MonthEnd(), freq='W-FRI')]
                settlement_date = fridays[1] # Get the second Friday
                settlement_dates.append(settlement_date)

                # Expiry date is the business day before settlement date
                expiry_dates.append(settlement_date - pd.offsets.BDay(1))

    df['expiry_date'] = expiry_dates
    df['settlement_date'] = settlement_dates

    # Calculate the effective period of the BBSW 3M fixing of the contract.
    df['effective_date'] = df['settlement_date']
    cal = get_busdaycal('AU-SYDNEY')
    df['maturity_date'] = workday(df['settlement_date'] + pd.DateOffset(months=3), 0, cal=cal)

    if 'price' in df.columns:
        # Per the ASX contract definition: Settlement Value = 100 - [BBSW 3M Fixing]
        # Hence: BBSW 3M Fixing implied by the price is 100 - price.
        df['futures_rate'] = (100 - df['price']) * 0.01

    return df





# -*- coding: utf-8 -*-
import numpy as np
from scipy.interpolate import interp1d
from frm.pricing_engine.monte_carlo_generic import generate_rand_nbs
from frm.utils import MAX_SIMULATIONS_PER_LOOP

def clewlow_strickland_1_factor_simulate(forward_curve, nb_simulations, segments_per_day, T, alpha, sigma):
    """
    Simulates spot prices using the Clewlow-Strickland one-factor model based on a provided forward curve.

    Parameters
    ----------
    forward_curve : numpy.ndarray
        A two-dimensional array where the first column contains time steps and the second column contains the corresponding forward prices.
    nb_simulations : int
        Number of simulation paths to generate.
    segments_per_day : int
        Number of extra simulation steps per forward curve step; higher values increase the model's temporal resolution.
    T : float
        Time horizon in days for the simulation.
    alpha : float
        Mean reversion speed of the model.
    sigma : float
        Volatility of the underlying asset.

    Returns
    -------
    spot_px_result : numpy.ndarray
        A two-dimensional array of simulated spot prices. Each row corresponds to a time step, and each column to a simulation path.
    
    Reference:
    Ahmos Sansom (2024). Clewlow and Strickland Commodity one factor spot model 
    (https://www.mathworks.com/matlabcentral/fileexchange/26969-clewlow-and-strickland-commodity-one-factor-spot-model), 
    MATLAB Central File Exchange. Retrieved April 28, 2024.        
    
    TODO
     Currently the code assumes the forward curve step has a constant increment.
     Want it to support a generic structure where the increment may change and may be denominated in days, months or years.
     The code simply discretizes it generally.
     Need to figure out if it should discretize evenly from t=0 to T=T or discretize based on the forward curve granularity (which may change)
     ...
     Probably easiest to simulate the entire forward curve, and not include the T input that Atmos did.
     The user can process the simulation results after.
     ...
     Likely the easiest way is to add extra granularity between the time-steps provided by the user.
     That gives the user more power as well - they can apply interpolation if they want rather than us assuming they want it.
     ...
     May not be relevant for this function but days is likely the best data structure for functions to expect your data.
     Easiest to convert to anything else given a year frac function 1st calls a day count function.
    """

    nb_simulations = int(nb_simulations)
    segments_per_day = int(segments_per_day)
    
    a = alpha
    σ = sigma

    # this would need to be adjusted to the time between t and t+1 based on the forward curve
    dt = 1.0 / segments_per_day 

    nb_steps = int(T * segments_per_day) # The Clewlow Strickland 1 Factor model requires a granular discretization in order to converge
    tau = (np.arange(nb_steps+1) * dt)
    
    f = interp1d(forward_curve[:,0], forward_curve[:,1], kind='linear') # Interpolate the forward curve based on the steps 
    forward_curve_interp = f(tau)
    index_spot = np.arange(T+1) * segments_per_day
    ln_forward_curve_interp = np.log(forward_curve_interp)
    
    nb_simulations_per_loop = int(min(nb_simulations, MAX_SIMULATIONS_PER_LOOP // nb_steps))
    nb_loops = (nb_simulations + nb_simulations_per_loop - 1) // nb_simulations_per_loop
    
    spot_px_result = np.zeros((T+1, nb_simulations))
    ln_spot_px = np.zeros((nb_steps+1, nb_simulations_per_loop))
    ln_spot_px[0, :] = ln_forward_curve_interp[0] # Set t=0 to the current spot price
    
    # These terms evaluate to constants hence can be calculated outside the monte carlo loop
    term1 = (ln_forward_curve_interp[1:] - ln_forward_curve_interp[:-1]) / dt
    term1 = np.concatenate(([np.nan], term1)) # Prefix with nan for t=0
    term3 = -0.25 * σ**2 * (1.0 + np.exp(-2.0 * a * tau[:-1]))
    term3 = np.concatenate(([np.nan], term3)) # Prefix with nan for t=0    

    for j in range(nb_loops):        
        rand_nbs = generate_rand_nbs(nb_steps=nb_steps, nb_rand_vars=1, nb_simulations=nb_simulations_per_loop, apply_antithetic_variates=False, random_seed=j)

        for i in range(1, ln_spot_px.shape[0]):
            term2 = a * (ln_forward_curve_interp[i-1] - ln_spot_px[i-1,:])
            ln_spot_px[i,:] = ln_spot_px[i-1,:] + (term1[i] + term2 + term3[i]) * dt + rand_nbs[i-1,0,:] * σ * np.sqrt(dt)   

        spot_px = np.exp(ln_spot_px)        
        
        idx_start = j*nb_simulations_per_loop
        idx_end = min((j+1)*nb_simulations_per_loop, nb_simulations)
        idx = np.arange(idx_start,idx_end)
        
        if j < (nb_loops-1):    
            spot_px_result[:, idx] = spot_px[index_spot, :]
        else:
            index_spot_2d = index_spot[:, np.newaxis]
            mask = np.arange(0,spot_px.shape[1]) < (idx_end-idx_start)
            spot_px_result[:, idx] = spot_px[index_spot_2d, mask]

    return spot_px_result
    


            
    











# -*- coding: utf-8 -*-
import os

if __name__ == "__main__":
    os.chdir(os.environ.get('PROJECT_DIR_FRM'))

from dataclasses import dataclass
import numpy as np
import pandas as pd

from frm.enums.utils import CompoundingFreq, PeriodFreq
from frm.enums.term_structures import OISCouponCalcMethod, TermRate
from frm.utils.daycount import year_frac
from frm.term_structures.zero_curve import ZeroCurve
from frm.term_structures.historical_swap_index_fixings import OISFixings, TermFixings



@dataclass
class TermSwapCurve:
    zero_curve: ZeroCurve
    historical_fixings: TermFixings

    # TODO: Term swap curve must require a fixing frequency.


    def __post_init__(self):
        assert self.zero_curve.day_count_basis == self.historical_fixings.day_count_basis

    def get_fixings(self,
                    period_start: pd.DatetimeIndex,
                    period_end: pd.DatetimeIndex,
                    fixing_freq: PeriodFreq, # TODO think about this and the enum
                    cpn_calc_method: CompoundingFreq=TermRate.SIMPLE):

        fixings = np.full(period_start.shape, np.nan)

        # TODO Choose the zero curve based on the fixing frequency
        #   Would be nice to have public API to call for data whenever needed for examples.
        # Get forward fixings
        mask_future = period_start >= self.zero_curve.curve_date
        forward_fixings = self.zero_curve.get_forward_rates(
            period_start=period_start[mask_future],
            period_end=period_end[mask_future],
            forward_rate_type=cpn_calc_method)
        fixings[mask_future] = forward_fixings

        # Get historical fixings
        mask_historical = period_start >= self.zero_curve.curve_date
        historical_fixings = self.historical_fixings.index_historical_fixings(
            fixing_dates=period_start[mask_historical])
        fixings[mask_historical] = historical_fixings

        return fixings


@dataclass
class OISCurve:
    zero_curve: ZeroCurve
    historical_fixings: OISFixings

    def __post_init__(self):
        assert self.zero_curve.day_count_basis == self.historical_fixings.day_count_basis

    def get_discount_factors(self, dates):
        return self.zero_curve.get_discount_factors(dates=dates)

    def get_fixings(self,
                    period_start: pd.DatetimeIndex,
                    period_end: pd.DatetimeIndex,
                    cpn_calc_method: OISCouponCalcMethod=OISCouponCalcMethod.DAILY_COMPOUNDED):

        fixings = np.full(period_start.shape, np.nan)

        # Get forward fixings
        mask_future = period_start >= self.zero_curve.curve_date
        if mask_future.any():
            forward_fixings = self.zero_curve.get_forward_rates(
                period_start=period_start[mask_future],
                period_end=period_end[mask_future],
                forward_rate_type=cpn_calc_method)
            fixings[mask_future] = forward_fixings

        # Get historical fixings
        mask_historical = period_end <= self.zero_curve.curve_date
        if mask_historical.any():
            historical_fixings = self.historical_fixings.get_coupon_rates(
                period_start=period_start[mask_historical],
                period_end=period_end[mask_historical],
                cpn_calc_method=cpn_calc_method)
            fixings[mask_historical] = historical_fixings

        # Calculate the current/cross-over fixing
        mask_cross_over = np.logical_and(period_start < self.zero_curve.curve_date, period_end > self.zero_curve.curve_date)
        if mask_cross_over.any():
            fixings[mask_cross_over] = self._get_crossover_coupon_rate(
                crossover_period_start=period_start[mask_cross_over],
                crossover_period_end=period_end[mask_cross_over],
                cpn_calc_method=cpn_calc_method)

        return fixings


    def _get_crossover_coupon_rate(self,
                                  crossover_period_start: pd.DatetimeIndex,
                                  crossover_period_end: pd.DatetimeIndex,
                                  cpn_calc_method: OISCouponCalcMethod):
        historical_year_frac = year_frac(crossover_period_start, self.zero_curve.curve_date, self.zero_curve.day_count_basis)
        future_year_frac = year_frac(self.zero_curve.curve_date, crossover_period_end, self.zero_curve.day_count_basis)

        curve_datetime_index = pd.DatetimeIndex([self.zero_curve.curve_date for _ in range(len(crossover_period_start))])

        historical_component = self.historical_fixings.get_coupon_rates(
            period_start=crossover_period_start,
            period_end=curve_datetime_index,
            cpn_calc_method=cpn_calc_method)
        future_component = self.zero_curve.get_forward_rates(
            period_start=curve_datetime_index,
            period_end=crossover_period_end,
            forward_rate_type=cpn_calc_method)

        crossover_rate = (historical_year_frac * historical_component
                          + (1 + historical_year_frac * historical_component) * future_year_frac * future_component) \
                         / (historical_year_frac + future_year_frac)

        return crossover_rate











# # -*- coding: utf-8 -*-
# import os
# if __name__ == "__main__":
#     os.chdir(os.environ.get('PROJECT_DIR_FRM'))
#
# from abc import ABC, abstractmethod
# from dataclasses import dataclass
# import numpy as np
# import pandas as pd
#
# from frm.enums.utils import DayCountBasis, OISCouponCalcMethod
#
#
# @dataclass
# class OISFixings:
#     fixings: pd.DataFrame
#     day_count_basis: DayCountBasis
#     ois_coupon_calc_method: OISCouponCalcMethod=None
#     name: str = None
#
#     def __post_init__(self):
#         required_columns = {'date', 'fixing'}
#         assert set(self.fixings.columns) == required_columns, \
#             f"Columns in fixings must be {required_columns}. Found: {set(self.fixings.columns)}"
#         assert isinstance(self.day_count_basis, DayCountBasis), \
#             "'day_count_basis' must be an instance of DayCountBasis"
#         if self.ois_coupon_calc_method is not None:
#             assert isinstance(self.ois_coupon_calc_method, OISCouponCalcMethod), \
#                 "'ois_coupon_calc_method' must be an instance of OISCouponCalcMethod"
#
#
#     def index_historical_fixings(self, fixing_dates: pd.DatetimeIndex):
#
#         # Additional Term-specific initialization if needed
#
#         # At a late date need to do functionality for short stub where coupon rate a calc over 1M and 3M bank bills.
#         # # Check for the required columns in fixings DataFrame
#         # required_columns = {'date'}
#         # missing_columns = required_columns - set(self.fixings.columns)
#         # assert not missing_columns, f"Missing columns in fixings: {missing_columns}"
#
#         # # Term fixings must be specified as 'fixing_{tenor}'
#         # # E.g: 'fixing_90d', 'fixing_3m', 'fixing_1y'
#         # # Can be multiple columns i.e 'fixing_30d', 'fixing_60d', 'fixing_90d'
#         # # Ensure all other columns follow the 'fixing_{tenor}' format
#         # invalid_columns = [col for col in self.fixings.columns if col not in required_columns and not col.startswith('fixing_')]
#         # assert not invalid_columns, f"Invalid columns found in fixings: {invalid_columns}"
#
#         # fixing_columns = [col for col in self.fixings.columns if col.startswith('fixing_')]
#         # tenors = [column_name.split('_')[-1] for column_name in fixing_columns]
#
#         # Create a DataFrame from the input dates and store the original index
#         date_df = pd.DataFrame({'date': fixing_dates})
#
#         # Sort the date DataFrame for the merge_asof operation
#         date_df_sorted = date_df.sort_values('date').reset_index(drop=False)
#
#         # Perform merge_asof with the sorted dates
#         df = pd.merge_asof(date_df_sorted, self.fixings, on='date', direction='backward')
#
#         # Restore the original order of the input dates
#         df_sorted_back = df.sort_values('index').set_index('index')
#
#         # Return the 'fixing' column in the order of the original input dates
#         return df_sorted_back['fixing'].values
#
#
#     def calc_historical_ois_coupon_rate(
#         self,
#         period_start: pd.DatetimeIndex,
#         period_end: pd.DatetimeIndex,
#         ois_coupon_calc_method: OISCouponCalcMethod = None,
#     ):
#
#         if ois_coupon_calc_method is None:
#             ois_coupon_calc_method = self.ois_coupon_calc_method
#
#         observations_start = period_start
#         observations_end = period_end - pd.DateOffset(days=1)
#
#         mask = np.logical_and(self.fixings['date'] >= observations_start.min(),
#                               self.fixings['date'] <= observations_end.max())
#         applicable_fixings = self.fixings.loc[mask,:]
#         applicable_fixings = applicable_fixings.sort_values('date', ascending=True).reset_index(drop=True)
#
#         if observations_end.max() > applicable_fixings['date'].max() \
#             or observations_end.min() < applicable_fixings['date'].min():
#             raise ValueError('Observation period extends beyond available fixing data.')
#
#         historical_coupon_rates = np.empty(shape=len(observations_start))
#         for i, (observation_start, observation_end) in enumerate(zip(observations_start, observations_end)):
#             mask = np.logical_and(applicable_fixings['date'] >= observation_start, applicable_fixings['date'] <= observation_end)
#
#             periods_ois_fixings = applicable_fixings.loc[mask,:]
#             periods_ois_fixings = periods_ois_fixings.sort_values('date', ascending=True).reset_index(drop=True)
#
#             if ois_coupon_calc_method == OISCouponCalcMethod.SIMPLE_AVERAGE:
#                 cpn_rate = periods_ois_fixings['fixing'].mean()
#             else:
#                 date_range = pd.date_range(start=observation_start, end=observation_end, freq='D')
#                 date_df = pd.DataFrame({'date': date_range})
#                 df = pd.merge_asof(date_df, periods_ois_fixings, on='date', direction='backward')
#                 df = df.sort_values('date', ascending=True).reset_index(drop=True)
#
#                 match ois_coupon_calc_method:
#                     case OISCouponCalcMethod.DAILY_COMPOUNDED:
#                         df['daily_interest'] = 1.0 + df['fixing'] / self.day_count_basis.days_per_year
#                         cpn_rate = (df['daily_interest'].prod() - 1.0) *  self.day_count_basis.days_per_year / len(date_range)
#                     case OISCouponCalcMethod.WEIGHTED_AVERAGE:
#                         cpn_rate = df['fixing'].mean()
#                     case _:
#                         raise ValueError(f"Invalid OISCouponCalcMethod: {ois_coupon_calc_method}")
#
#             historical_coupon_rates[i] = cpn_rate
#
#         return historical_coupon_rates
#
#
#